from __future__ import annotations

import argparse
import inspect
import json
import math
from collections import Counter
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Any

import msgspec
from grim.geom import Vec2

import crimson.projectiles as projectiles_mod
import crimson.sim.presentation_step as presentation_step_mod
from crimson.game_modes import GameMode
from crimson.gameplay import PlayerInput
from crimson.original.capture import (
    CAPTURE_BOOTSTRAP_EVENT_KIND,
    build_capture_dt_frame_overrides,
    build_capture_dt_frame_ms_i32_overrides,
    capture_bootstrap_payload_from_event_payload,
    convert_capture_to_replay,
    load_capture,
)
from crimson.replay.types import UnknownEvent, unpack_input_flags, unpack_packed_player_input
from crimson.sim.runners.common import (
    build_damage_scale_by_type,
    build_empty_fx_queues,
    reset_players,
    status_from_snapshot,
)
from crimson.sim.runners.survival import (
    _apply_tick_events,
    _decode_digital_move_keys,
    _partition_tick_events,
    _resolve_dt_frame,
)
from crimson.sim.sessions import SurvivalDeterministicSession
from crimson.sim.world_state import WorldState


@dataclass(slots=True)
class CollisionRow:
    proj_index: int | None
    proj_type: int | None
    proj_life: float | None
    step: int | None
    creature_idx: int | None
    margin: float
    dist: float
    radius: float
    threshold: float
    hit: bool


@dataclass(slots=True)
class FocusTraceReport:
    tick: int
    hits: int
    deaths: int
    sfx: int
    rand_calls_total: int
    rng_callsites_top: list[tuple[str, int]]
    rng_callsites_head: list[str]
    collision_hits: list[CollisionRow]
    collision_near_misses: list[CollisionRow]
    pre_projectiles: list[dict[str, Any]]
    post_projectiles: list[dict[str, Any]]
    capture_projectiles: list[dict[str, Any]]
    capture_creatures: list[dict[str, Any]]
    creature_diffs_top: list[dict[str, Any]]
    creature_capture_only: list[dict[str, Any]]
    creature_rewrite_only: list[dict[str, Any]]
    projectile_diffs_top: list[dict[str, Any]]
    projectile_capture_only: list[dict[str, Any]]
    projectile_rewrite_only: list[dict[str, Any]]
    decal_hook_rows: list[DecalHookRow]
    rng_alignment: RngAlignmentSummary
    native_caller_gaps_top: list[NativeCallerGapRow]
    fire_bullets_loop_parity: FireBulletsLoopParity | None


@dataclass(slots=True)
class DecalHookRow:
    hook_index: int
    type_id: int
    handled: bool
    rng_draws: int
    target_x: float
    target_y: float


@dataclass(slots=True)
class RngAlignmentTailRow:
    index: int
    capture_value: int
    capture_caller_static: str
    capture_caller: str
    inferred_rewrite_callsite: str


@dataclass(slots=True)
class RngAlignmentSummary:
    capture_calls: int
    capture_head_len: int
    rewrite_calls: int
    value_prefix_match: int
    first_value_mismatch_index: int | None
    first_value_mismatch_capture: int | None
    first_value_mismatch_rewrite: int | None
    missing_native_tail_count: int
    missing_native_tail_callers_top: list[tuple[str, int]]
    missing_native_tail_inferred_callsites_top: list[tuple[str, int]]
    missing_native_tail_preview: list[RngAlignmentTailRow]
    capture_caller_counts: list[tuple[str, int]]
    rewrite_callsite_counts: list[tuple[str, int]]
    caller_static_to_rewrite_callsite: list[tuple[str, str]]


@dataclass(slots=True)
class NativeCallerGapRow:
    native_caller_static: str
    native_label: str
    capture_count: int
    inferred_rewrite_callsite: str
    rewrite_count: int
    gap: int


@dataclass(slots=True)
class FireBulletsLoopParity:
    capture_iterations: int
    rewrite_iterations: int
    missing_iterations: int
    loop_iterations_per_hit: int
    estimated_missing_hits: float
    capture_midrange_rolls: int
    rewrite_midrange_rolls: int
    capture_farrange_rolls: int
    rewrite_farrange_rolls: int
    capture_pre_freeze_rolls: int
    rewrite_pre_freeze_rolls: int


_NATIVE_CALLER_LABELS: dict[str, str] = {
    "0x0042176f": "projectile_update.fire_bullets_loop_seed",
    "0x00421799": "projectile_update.fire_bullets_midrange_reroll",
    "0x004217c6": "projectile_update.fire_bullets_farrange_reroll",
    "0x0042184c": "projectile_update.fire_bullets_pre_freeze_rand",
    "0x00427760": "fx_queue_add_random.gray_tint_rand",
    "0x0042778e": "fx_queue_add_random.size_rand",
    "0x004277b0": "fx_queue_add_random.rotation_rand",
    "0x0042780b": "fx_queue_add_random.effect_id_rand",
    "0x0042ebc0": "effect_spawn_blood_splatter.rotation_rand",
    "0x0042ebe3": "effect_spawn_blood_splatter.half_size_rand",
    "0x0042ec00": "effect_spawn_blood_splatter.speed_x_rand",
    "0x0042ec1d": "effect_spawn_blood_splatter.speed_y_rand",
    "0x0042ec44": "effect_spawn_blood_splatter.scale_step_rand",
}

_FIRE_BULLETS_SEED_CALLER = "0x0042176f"
_FIRE_BULLETS_MIDRANGE_CALLER = "0x00421799"
_FIRE_BULLETS_FARRANGE_CALLER = "0x004217c6"
_FIRE_BULLETS_PRE_FREEZE_CALLER = "0x0042184c"
_FIRE_BULLETS_LOOP_ITERS_PER_HIT = 6
_JSON_OUT_AUTO = "__AUTO__"
_DEFAULT_JSON_OUT_DIR = Path("artifacts/frida/reports")


def _resolve_json_out_path(value: str | None, *, tick: int) -> Path | None:
    if value is None:
        return None
    if str(value) == _JSON_OUT_AUTO:
        return _DEFAULT_JSON_OUT_DIR / f"focus_trace_tick{int(tick)}_latest.json"
    return Path(value)


def _read_capture_tick(capture: object, tick: int) -> dict[str, Any] | None:
    ticks = getattr(capture, "ticks", None)
    if not isinstance(ticks, list):
        return None
    for row in ticks:
        if int(getattr(row, "tick_index", -1)) != int(tick):
            continue
        obj = msgspec.to_builtins(row)
        if isinstance(obj, dict):
            return obj
    return None


def _projectile_snapshot(world: WorldState) -> list[dict[str, Any]]:
    out: list[dict[str, Any]] = []
    for idx, proj in enumerate(world.state.projectiles.entries):
        if not bool(proj.active):
            continue
        out.append(
            {
                "index": int(idx),
                "type_id": int(proj.type_id),
                "life_timer": float(proj.life_timer),
                "damage_pool": float(proj.damage_pool),
                "hit_radius": float(proj.hit_radius),
                "base_damage": float(proj.base_damage),
                "owner_id": int(proj.owner_id),
                "pos": {"x": float(proj.pos.x), "y": float(proj.pos.y)},
            }
        )
    return out


def _decode_inputs_for_tick(
    replay: Any,
    tick_index: int,
    *,
    original_capture_replay: bool,
    digital_move_enabled_by_player: set[int],
) -> list[PlayerInput]:
    packed_tick = replay.inputs[int(tick_index)]
    out: list[PlayerInput] = []
    for player_index, packed in enumerate(packed_tick):
        mx, my, ax, ay, flags = unpack_packed_player_input(packed)
        fire_down, fire_pressed, reload_pressed = unpack_input_flags(int(flags))
        move_forward_pressed: bool | None = None
        move_backward_pressed: bool | None = None
        turn_left_pressed: bool | None = None
        turn_right_pressed: bool | None = None
        if original_capture_replay and int(player_index) in digital_move_enabled_by_player:
            digital_move = _decode_digital_move_keys(float(mx), float(my))
            if digital_move is not None:
                (
                    move_forward_pressed,
                    move_backward_pressed,
                    turn_left_pressed,
                    turn_right_pressed,
                ) = digital_move
        out.append(
            PlayerInput(
                move=Vec2(float(mx), float(my)),
                aim=Vec2(float(ax), float(ay)),
                fire_down=bool(fire_down),
                fire_pressed=bool(fire_pressed),
                reload_pressed=bool(reload_pressed),
                move_forward_pressed=move_forward_pressed,
                move_backward_pressed=move_backward_pressed,
                turn_left_pressed=turn_left_pressed,
                turn_right_pressed=turn_right_pressed,
            )
        )
    return out


def _load_capture_events(replay: Any) -> tuple[dict[int, list[object]], bool, set[int]]:
    events_by_tick: dict[int, list[object]] = {}
    original_capture_replay = False
    digital_move_enabled_by_player: set[int] = set()
    for event in replay.events:
        if isinstance(event, UnknownEvent) and str(event.kind) == CAPTURE_BOOTSTRAP_EVENT_KIND:
            original_capture_replay = True
            payload = capture_bootstrap_payload_from_event_payload(list(event.payload))
            if isinstance(payload, dict):
                enabled_raw = payload.get("digital_move_enabled_by_player")
                if isinstance(enabled_raw, list):
                    for player_index, enabled in enumerate(enabled_raw):
                        if bool(enabled):
                            digital_move_enabled_by_player.add(int(player_index))
        events_by_tick.setdefault(int(event.tick_index), []).append(event)
    return events_by_tick, original_capture_replay, digital_move_enabled_by_player


def _summarize_creature_diffs(capture_creatures: list[dict[str, Any]], world: WorldState) -> list[dict[str, Any]]:
    cap_by_idx: dict[int, dict[str, Any]] = {
        int(row.get("index")): row for row in capture_creatures if isinstance(row, dict) and row.get("index") is not None  # ty:ignore[invalid-argument-type]
    }
    rows: list[dict[str, Any]] = []
    for idx, cap_row in cap_by_idx.items():
        if not (0 <= int(idx) < len(world.creatures.entries)):
            continue
        creature = world.creatures.entries[int(idx)]
        cap_pos = cap_row.get("pos") if isinstance(cap_row.get("pos"), dict) else {}
        cap_x = float(cap_pos.get("x", 0.0))  # ty:ignore[possibly-missing-attribute]
        cap_y = float(cap_pos.get("y", 0.0))  # ty:ignore[possibly-missing-attribute]
        cap_hp = float(cap_row.get("hp", 0.0))
        cap_hitbox = float(cap_row.get("hitbox_size", 0.0))
        rows.append(
            {
                "index": int(idx),
                "hp_delta": float(creature.hp) - cap_hp,
                "hitbox_delta": float(creature.hitbox_size) - cap_hitbox,
                "x_delta": float(creature.pos.x) - cap_x,
                "y_delta": float(creature.pos.y) - cap_y,
                "active_capture": bool(int(cap_row.get("active", 0)) != 0),
                "active_rewrite": bool(creature.active),
            }
        )
    rows.sort(
        key=lambda row: (
            -max(
                abs(float(row["hp_delta"])),
                abs(float(row["hitbox_delta"])),
                abs(float(row["x_delta"])),
                abs(float(row["y_delta"])),
            ),
            int(row["index"]),
        )
    )
    return rows


def _summarize_projectile_diffs(capture_projectiles: list[dict[str, Any]], world: WorldState) -> list[dict[str, Any]]:
    cap_by_idx: dict[int, dict[str, Any]] = {
        int(row.get("index")): row for row in capture_projectiles if isinstance(row, dict) and row.get("index") is not None  # ty:ignore[invalid-argument-type]
    }
    rows: list[dict[str, Any]] = []
    for idx, cap_row in cap_by_idx.items():
        if not (0 <= int(idx) < len(world.state.projectiles.entries)):
            continue
        proj = world.state.projectiles.entries[int(idx)]
        cap_pos = cap_row.get("pos") if isinstance(cap_row.get("pos"), dict) else {}
        cap_x = float(cap_pos.get("x", 0.0))  # ty:ignore[possibly-missing-attribute]
        cap_y = float(cap_pos.get("y", 0.0))  # ty:ignore[possibly-missing-attribute]
        rows.append(
            {
                "index": int(idx),
                "life_delta": float(proj.life_timer) - float(cap_row.get("life_timer", 0.0)),
                "damage_pool_delta": float(proj.damage_pool) - float(cap_row.get("damage_pool", 0.0)),
                "x_delta": float(proj.pos.x) - cap_x,
                "y_delta": float(proj.pos.y) - cap_y,
                "active_capture": bool(int(cap_row.get("active", 0)) != 0),
                "active_rewrite": bool(proj.active),
            }
        )
    rows.sort(
        key=lambda row: (
            -max(
                abs(float(row["life_delta"])),
                abs(float(row["damage_pool_delta"])),
                abs(float(row["x_delta"])),
                abs(float(row["y_delta"])),
            ),
            int(row["index"]),
        )
    )
    return rows


def _collect_creature_presence_diffs(
    capture_creatures: list[dict[str, Any]],
    world: WorldState,
) -> tuple[list[dict[str, Any]], list[dict[str, Any]]]:
    cap_by_idx: dict[int, dict[str, Any]] = {
        int(row.get("index")): row for row in capture_creatures if isinstance(row, dict) and row.get("index") is not None  # ty:ignore[invalid-argument-type]
    }
    cap_indices = {idx for idx, row in cap_by_idx.items() if bool(int(row.get("active", 0)) != 0)}
    rewrite_indices = {idx for idx, creature in enumerate(world.creatures.entries) if bool(creature.active)}

    capture_only: list[dict[str, Any]] = []
    for idx in sorted(cap_indices - rewrite_indices):
        row = cap_by_idx[int(idx)]
        pos = row.get("pos") if isinstance(row.get("pos"), dict) else {}
        capture_only.append(
            {
                "index": int(idx),
                "type_id": int(row.get("type_id", 0)),
                "hp": float(row.get("hp", 0.0)),
                "hitbox_size": float(row.get("hitbox_size", 0.0)),
                "pos": {"x": float(pos.get("x", 0.0)), "y": float(pos.get("y", 0.0))},  # ty:ignore[possibly-missing-attribute]
            }
        )

    rewrite_only: list[dict[str, Any]] = []
    for idx in sorted(rewrite_indices - cap_indices):
        creature = world.creatures.entries[int(idx)]
        rewrite_only.append(
            {
                "index": int(idx),
                "type_id": int(creature.type_id),
                "hp": float(creature.hp),
                "hitbox_size": float(creature.hitbox_size),
                "pos": {"x": float(creature.pos.x), "y": float(creature.pos.y)},
            }
        )

    return capture_only, rewrite_only


def _collect_projectile_presence_diffs(
    capture_projectiles: list[dict[str, Any]],
    world: WorldState,
) -> tuple[list[dict[str, Any]], list[dict[str, Any]]]:
    cap_by_idx: dict[int, dict[str, Any]] = {
        int(row.get("index")): row for row in capture_projectiles if isinstance(row, dict) and row.get("index") is not None  # ty:ignore[invalid-argument-type]
    }
    cap_indices = {idx for idx, row in cap_by_idx.items() if bool(int(row.get("active", 0)) != 0)}
    rewrite_indices = {idx for idx, proj in enumerate(world.state.projectiles.entries) if bool(proj.active)}

    capture_only: list[dict[str, Any]] = []
    for idx in sorted(cap_indices - rewrite_indices):
        row = cap_by_idx[int(idx)]
        pos = row.get("pos") if isinstance(row.get("pos"), dict) else {}
        capture_only.append(
            {
                "index": int(idx),
                "type_id": int(row.get("type_id", 0)),
                "life_timer": float(row.get("life_timer", 0.0)),
                "damage_pool": float(row.get("damage_pool", 0.0)),
                "pos": {"x": float(pos.get("x", 0.0)), "y": float(pos.get("y", 0.0))},  # ty:ignore[possibly-missing-attribute]
            }
        )

    rewrite_only: list[dict[str, Any]] = []
    for idx in sorted(rewrite_indices - cap_indices):
        proj = world.state.projectiles.entries[int(idx)]
        rewrite_only.append(
            {
                "index": int(idx),
                "type_id": int(proj.type_id),
                "life_timer": float(proj.life_timer),
                "damage_pool": float(proj.damage_pool),
                "pos": {"x": float(proj.pos.x), "y": float(proj.pos.y)},
            }
        )

    return capture_only, rewrite_only


def _summarize_rng_alignment(
    *,
    capture_rng_head: list[dict[str, Any]],
    capture_rng_calls: int,
    rewrite_rng_values: list[int],
    rewrite_rng_callsites: list[str],
    tail_preview_limit: int = 24,
) -> RngAlignmentSummary:
    capture_values = [int(row.get("value", 0)) for row in capture_rng_head if isinstance(row, dict)]
    capture_caller_counts = Counter(
        str(row.get("caller_static", "")).strip()
        for row in capture_rng_head
        if isinstance(row, dict) and str(row.get("caller_static", "")).strip()
    )
    rewrite_callsite_counts = Counter(str(callsite) for callsite in rewrite_rng_callsites if str(callsite))
    capture_calls = max(int(capture_rng_calls), len(capture_values))
    rewrite_calls = len(rewrite_rng_values)
    min_len = min(len(capture_values), rewrite_calls)

    prefix = 0
    while prefix < min_len:
        if int(capture_values[prefix]) != int(rewrite_rng_values[prefix]):
            break
        prefix += 1

    mismatch_index: int | None = None
    mismatch_capture: int | None = None
    mismatch_rewrite: int | None = None
    if prefix < min_len:
        mismatch_index = int(prefix)
        mismatch_capture = int(capture_values[prefix])
        mismatch_rewrite = int(rewrite_rng_values[prefix])

    caller_to_rewrite: dict[str, Counter[str]] = {}
    aligned = min(prefix, len(rewrite_rng_callsites), len(capture_rng_head))
    for idx in range(aligned):
        row = capture_rng_head[idx]
        if not isinstance(row, dict):
            continue
        caller_static = str(row.get("caller_static", "")).strip()
        if not caller_static:
            continue
        bucket = caller_to_rewrite.setdefault(caller_static, Counter())
        bucket[str(rewrite_rng_callsites[idx])] += 1

    caller_best: dict[str, str] = {}
    for caller_static, counts in caller_to_rewrite.items():
        top = counts.most_common(1)
        if not top:
            continue
        caller_best[str(caller_static)] = str(top[0][0])

    tail_start = min(rewrite_calls, len(capture_rng_head))
    tail_rows_raw = capture_rng_head[tail_start:]
    tail_callers = Counter(
        str(row.get("caller_static", "")).strip()
        for row in tail_rows_raw
        if isinstance(row, dict) and str(row.get("caller_static", "")).strip()
    )
    tail_inferred_callsites = Counter(
        str(caller_best.get(caller_static, "<unknown>"))
        for row in tail_rows_raw
        if isinstance(row, dict)
        for caller_static in [str(row.get("caller_static", "")).strip()]
        if caller_static
    )
    tail_preview: list[RngAlignmentTailRow] = []
    for offset, row in enumerate(tail_rows_raw[: max(0, int(tail_preview_limit))]):
        if not isinstance(row, dict):
            continue
        caller_static = str(row.get("caller_static", "")).strip()
        tail_preview.append(
            RngAlignmentTailRow(
                index=int(tail_start + offset),
                capture_value=int(row.get("value", 0)),
                capture_caller_static=caller_static,
                capture_caller=str(row.get("caller", "")).strip(),
                inferred_rewrite_callsite=str(caller_best.get(caller_static, "")),
            )
        )

    missing_native_tail_count = max(0, int(capture_calls) - int(rewrite_calls))
    return RngAlignmentSummary(
        capture_calls=int(capture_calls),
        capture_head_len=int(len(capture_values)),
        rewrite_calls=int(rewrite_calls),
        value_prefix_match=int(prefix),
        first_value_mismatch_index=mismatch_index,
        first_value_mismatch_capture=mismatch_capture,
        first_value_mismatch_rewrite=mismatch_rewrite,
        missing_native_tail_count=int(missing_native_tail_count),
        missing_native_tail_callers_top=[(str(key), int(count)) for key, count in tail_callers.most_common(12)],
        missing_native_tail_inferred_callsites_top=[
            (str(key), int(count)) for key, count in tail_inferred_callsites.most_common(12)
        ],
        missing_native_tail_preview=tail_preview,
        capture_caller_counts=[(str(key), int(count)) for key, count in capture_caller_counts.most_common()],
        rewrite_callsite_counts=[(str(key), int(count)) for key, count in rewrite_callsite_counts.most_common()],
        caller_static_to_rewrite_callsite=sorted(
            ((str(caller_static), str(rewrite_callsite)) for caller_static, rewrite_callsite in caller_best.items()),
            key=lambda item: item[0],
        ),
    )


def _build_native_caller_gaps(
    rng_alignment: RngAlignmentSummary,
    *,
    limit: int = 20,
) -> list[NativeCallerGapRow]:
    caller_map = {
        str(caller_static): str(rewrite_callsite)
        for caller_static, rewrite_callsite in rng_alignment.caller_static_to_rewrite_callsite
    }
    rewrite_counts = Counter(
        {
            str(callsite): int(count)
            for callsite, count in rng_alignment.rewrite_callsite_counts
            if str(callsite)
        }
    )
    rows: list[NativeCallerGapRow] = []
    for caller_static, capture_count in rng_alignment.capture_caller_counts:
        native_caller_static = str(caller_static)
        if not native_caller_static:
            continue
        inferred_rewrite_callsite = str(caller_map.get(native_caller_static, ""))
        rewrite_count = int(rewrite_counts.get(inferred_rewrite_callsite, 0)) if inferred_rewrite_callsite else 0
        gap = int(capture_count) - int(rewrite_count)
        if gap <= 0:
            continue
        rows.append(
            NativeCallerGapRow(
                native_caller_static=native_caller_static,
                native_label=str(_NATIVE_CALLER_LABELS.get(native_caller_static, "")),
                capture_count=int(capture_count),
                inferred_rewrite_callsite=inferred_rewrite_callsite,
                rewrite_count=int(rewrite_count),
                gap=int(gap),
            )
        )
    rows.sort(
        key=lambda row: (
            -int(row.gap),
            -int(row.capture_count),
            str(row.native_caller_static),
        )
    )
    return rows[: max(0, int(limit))]


def _build_fire_bullets_loop_parity(rng_alignment: RngAlignmentSummary) -> FireBulletsLoopParity | None:
    caller_map = {
        str(caller_static): str(rewrite_callsite)
        for caller_static, rewrite_callsite in rng_alignment.caller_static_to_rewrite_callsite
    }
    capture_counts = Counter(
        {
            str(caller_static): int(count)
            for caller_static, count in rng_alignment.capture_caller_counts
            if str(caller_static)
        }
    )
    rewrite_counts = Counter(
        {
            str(callsite): int(count)
            for callsite, count in rng_alignment.rewrite_callsite_counts
            if str(callsite)
        }
    )

    seed_callsite = str(caller_map.get(_FIRE_BULLETS_SEED_CALLER, ""))
    seed_capture = int(capture_counts.get(_FIRE_BULLETS_SEED_CALLER, 0))
    seed_rewrite = int(rewrite_counts.get(seed_callsite, 0)) if seed_callsite else 0

    pre_freeze_callsite = str(caller_map.get(_FIRE_BULLETS_PRE_FREEZE_CALLER, ""))
    pre_freeze_capture = int(capture_counts.get(_FIRE_BULLETS_PRE_FREEZE_CALLER, 0))
    pre_freeze_rewrite = int(rewrite_counts.get(pre_freeze_callsite, 0)) if pre_freeze_callsite else 0

    midrange_callsite = str(caller_map.get(_FIRE_BULLETS_MIDRANGE_CALLER, ""))
    midrange_capture = int(capture_counts.get(_FIRE_BULLETS_MIDRANGE_CALLER, 0))
    midrange_rewrite = int(rewrite_counts.get(midrange_callsite, 0)) if midrange_callsite else 0

    farrange_callsite = str(caller_map.get(_FIRE_BULLETS_FARRANGE_CALLER, ""))
    farrange_capture = int(capture_counts.get(_FIRE_BULLETS_FARRANGE_CALLER, 0))
    farrange_rewrite = int(rewrite_counts.get(farrange_callsite, 0)) if farrange_callsite else 0

    if (
        seed_capture <= 0
        and seed_rewrite <= 0
        and pre_freeze_capture <= 0
        and pre_freeze_rewrite <= 0
        and midrange_capture <= 0
        and midrange_rewrite <= 0
        and farrange_capture <= 0
        and farrange_rewrite <= 0
    ):
        return None

    missing_iterations = max(0, int(seed_capture) - int(seed_rewrite))
    estimated_missing_hits = float(missing_iterations) / float(_FIRE_BULLETS_LOOP_ITERS_PER_HIT)
    return FireBulletsLoopParity(
        capture_iterations=int(seed_capture),
        rewrite_iterations=int(seed_rewrite),
        missing_iterations=int(missing_iterations),
        loop_iterations_per_hit=int(_FIRE_BULLETS_LOOP_ITERS_PER_HIT),
        estimated_missing_hits=float(estimated_missing_hits),
        capture_midrange_rolls=int(midrange_capture),
        rewrite_midrange_rolls=int(midrange_rewrite),
        capture_farrange_rolls=int(farrange_capture),
        rewrite_farrange_rolls=int(farrange_rewrite),
        capture_pre_freeze_rolls=int(pre_freeze_capture),
        rewrite_pre_freeze_rolls=int(pre_freeze_rewrite),
    )


def trace_focus_tick(
    *,
    capture_path: Path,
    tick: int,
    near_miss_threshold: float,
    inter_tick_rand_draws: int,
) -> FocusTraceReport:
    capture = load_capture(capture_path)
    replay = convert_capture_to_replay(capture)
    mode = int(replay.header.game_mode_id)
    if mode != int(GameMode.SURVIVAL):
        raise ValueError(f"focus trace currently supports survival mode only (got mode={mode})")

    raw_tick = _read_capture_tick(capture, int(tick))
    if raw_tick is None:
        raise ValueError(f"capture tick {tick} not found in {capture_path}")
    samples = raw_tick.get("samples") if isinstance(raw_tick.get("samples"), dict) else {}
    capture_creatures = samples.get("creatures") if isinstance(samples.get("creatures"), list) else []  # ty:ignore[possibly-missing-attribute]
    capture_projectiles = samples.get("projectiles") if isinstance(samples.get("projectiles"), list) else []  # ty:ignore[possibly-missing-attribute]
    capture_rng = raw_tick.get("rng") if isinstance(raw_tick.get("rng"), dict) else {}
    capture_rng_head = capture_rng.get("head") if isinstance(capture_rng.get("head"), list) else []  # ty:ignore[possibly-missing-attribute]
    capture_rng_calls = int(capture_rng.get("calls", len(capture_rng_head)))  # ty:ignore[invalid-argument-type, possibly-missing-attribute]

    world_size = float(replay.header.world_size)
    world = WorldState.build(
        world_size=world_size,
        demo_mode_active=False,
        hardcore=bool(replay.header.hardcore),
        difficulty_level=int(replay.header.difficulty_level),
        preserve_bugs=bool(replay.header.preserve_bugs),
    )
    reset_players(world.players, world_size=world_size, player_count=int(replay.header.player_count))
    world.state.status = status_from_snapshot(
        quest_unlock_index=int(replay.header.status.quest_unlock_index),
        quest_unlock_index_full=int(replay.header.status.quest_unlock_index_full),
        weapon_usage_counts=replay.header.status.weapon_usage_counts,
    )
    world.state.rng.srand(int(replay.header.seed))

    fx_queue, fx_queue_rotated = build_empty_fx_queues()
    session = SurvivalDeterministicSession(
        world=world,
        world_size=world_size,
        damage_scale_by_type=build_damage_scale_by_type(),
        fx_queue=fx_queue,
        fx_queue_rotated=fx_queue_rotated,
        detail_preset=5,
        fx_toggle=0,
        game_tune_started=False,
        clear_fx_queues_each_tick=True,
    )

    events_by_tick, original_capture_replay, digital_move_enabled_by_player = _load_capture_events(replay)
    dt_frame_overrides = build_capture_dt_frame_overrides(capture, tick_rate=int(replay.header.tick_rate))
    dt_frame_ms_i32_overrides = build_capture_dt_frame_ms_i32_overrides(capture)
    default_dt_frame = 1.0 / float(int(replay.header.tick_rate))
    outside_draws_by_tick = {
        int(item.tick_index): int(item.rng.outside_before_calls)
        for item in capture.ticks
        if int(item.rng.outside_before_calls) >= 0
    }
    if outside_draws_by_tick:
        first_tick_index = min(outside_draws_by_tick)
        # The inferred replay seed already matches the first sampled capture tick.
        outside_draws_by_tick[int(first_tick_index)] = 0
    use_outside_draws = bool(outside_draws_by_tick)

    rng_callsites: Counter[str] = Counter()
    rng_head: list[str] = []
    rng_values: list[int] = []
    rng_values_callsites: list[str] = []
    collision_hits: list[CollisionRow] = []
    near_misses: list[CollisionRow] = []
    decal_hook_rows: list[DecalHookRow] = []
    pre_projectiles: list[dict[str, Any]] = []
    post_projectiles: list[dict[str, Any]] = []
    focus_hits = 0
    focus_deaths = 0
    focus_sfx = 0

    root = Path.cwd().resolve()
    orig_rand = world.state.rng.rand
    orig_particles_rand = world.state.particles._rand
    orig_sprite_effects_rand = world.state.sprite_effects._rand
    orig_within = projectiles_mod._within_native_find_radius
    orig_run_projectile_decal_hooks = presentation_step_mod.run_projectile_decal_hooks

    try:
        for tick_index in range(int(tick) + 1):
            world.state.game_mode = int(GameMode.SURVIVAL)
            world.state.demo_mode_active = False
            if use_outside_draws:
                draws = outside_draws_by_tick.get(int(tick_index))
                if draws is None:
                    draws = int(inter_tick_rand_draws)
                for _ in range(max(0, int(draws))):
                    world.state.rng.rand()
            dt_tick = _resolve_dt_frame(
                tick_index=int(tick_index),
                default_dt_frame=float(default_dt_frame),
                dt_frame_overrides=dt_frame_overrides,
            )
            dt_tick_ms_i32 = dt_frame_ms_i32_overrides.get(int(tick_index))
            tick_events = events_by_tick.get(int(tick_index), [])
            pre_step_events, post_step_events = _partition_tick_events(
                tick_events,
                defer_menu_open=bool(original_capture_replay),
            )

            _apply_tick_events(
                pre_step_events,
                tick_index=int(tick_index),
                dt_frame=float(dt_tick),
                world=world,
                strict_events=False,
            )
            player_inputs = _decode_inputs_for_tick(
                replay,
                int(tick_index),
                original_capture_replay=bool(original_capture_replay),
                digital_move_enabled_by_player=digital_move_enabled_by_player,
            )

            if int(tick_index) == int(tick):
                pre_projectiles = _projectile_snapshot(world)

                def traced_rand() -> int:
                    value = int(orig_rand())
                    frame = inspect.currentframe()
                    caller = frame.f_back if frame is not None else None
                    key = "<unknown>"
                    while caller is not None:
                        filename = Path(caller.f_code.co_filename).resolve()
                        try:
                            rel = filename.relative_to(root)
                        except ValueError:
                            rel = filename
                        rel_s = str(rel)
                        if "src/crimson/" in rel_s:
                            key = f"{rel_s}:{caller.f_code.co_name}:{caller.f_lineno}"
                            break
                        caller = caller.f_back
                    rng_callsites[key] += 1
                    if len(rng_head) < 256:
                        rng_head.append(key)
                    rng_values.append(int(value))
                    rng_values_callsites.append(str(key))
                    return value

                def traced_within_native_find_radius(
                    *,
                    origin: Vec2,
                    target: Vec2,
                    radius: float,
                    target_size: float,
                ) -> bool:
                    dx = float(target.x) - float(origin.x)
                    dy = float(target.y) - float(origin.y)
                    dist = math.sqrt(dx * dx + dy * dy)
                    threshold = float(target_size) * 0.14285715 + 3.0
                    margin = dist - float(radius) - threshold
                    hit = bool(margin < 0.0)
                    frame = inspect.currentframe().f_back  # ty:ignore[possibly-missing-attribute]
                    proj_index: int | None = None
                    proj_type: int | None = None
                    proj_life: float | None = None
                    step: int | None = None
                    creature_idx: int | None = None
                    if frame is not None:
                        try:
                            step = int(frame.f_locals.get("step")) if "step" in frame.f_locals else None  # ty:ignore[invalid-argument-type]
                        except Exception:
                            step = None
                        try:
                            creature_idx = int(frame.f_locals.get("idx")) if "idx" in frame.f_locals else None  # ty:ignore[invalid-argument-type]
                        except Exception:
                            creature_idx = None
                        try:
                            proj_index = int(frame.f_locals.get("proj_index")) if "proj_index" in frame.f_locals else None  # ty:ignore[invalid-argument-type]
                        except Exception:
                            proj_index = None
                        proj = frame.f_locals.get("proj")
                        if proj is not None:
                            try:
                                proj_type = int(getattr(proj, "type_id"))
                                proj_life = float(getattr(proj, "life_timer"))
                            except Exception:
                                proj_type = None
                                proj_life = None
                    row = CollisionRow(
                        proj_index=proj_index,
                        proj_type=proj_type,
                        proj_life=proj_life,
                        step=step,
                        creature_idx=creature_idx,
                        margin=float(margin),
                        dist=float(dist),
                        radius=float(radius),
                        threshold=float(threshold),
                        hit=bool(hit),
                    )
                    if bool(hit):
                        collision_hits.append(row)
                    elif 0.0 <= float(margin) <= float(near_miss_threshold):
                        near_misses.append(row)
                    return bool(hit)

                hook_index = 0

                def traced_run_projectile_decal_hooks(ctx: Any) -> bool:
                    nonlocal hook_index
                    before = len(rng_values)
                    handled = bool(orig_run_projectile_decal_hooks(ctx))
                    after = len(rng_values)
                    hit = ctx.hit
                    decal_hook_rows.append(
                        DecalHookRow(
                            hook_index=int(hook_index),
                            type_id=int(hit.type_id),
                            handled=bool(handled),
                            rng_draws=max(0, int(after - before)),
                            target_x=float(hit.target.x),
                            target_y=float(hit.target.y),
                        )
                    )
                    hook_index += 1
                    return bool(handled)

                world.state.rng.rand = traced_rand  # type: ignore[assignment]
                world.state.particles._rand = traced_rand
                world.state.sprite_effects._rand = traced_rand
                projectiles_mod._within_native_find_radius = traced_within_native_find_radius  # type: ignore[assignment]
                presentation_step_mod.run_projectile_decal_hooks = traced_run_projectile_decal_hooks  # type: ignore[assignment]

            tick_result = session.step_tick(
                dt_frame=float(dt_tick),
                dt_frame_ms_i32=(int(dt_tick_ms_i32) if dt_tick_ms_i32 is not None else None),
                inputs=player_inputs,
                trace_rng=False,
            )

            if int(tick_index) == int(tick):
                post_projectiles = _projectile_snapshot(world)
                focus_hits = int(len(tick_result.step.events.hits))
                focus_deaths = int(len(tick_result.step.events.deaths))
                focus_sfx = int(len(tick_result.step.events.sfx))

            if post_step_events:
                _apply_tick_events(
                    post_step_events,
                    tick_index=int(tick_index),
                    dt_frame=float(dt_tick),
                    world=world,
                    strict_events=False,
                )

            if int(tick_index) == int(tick):
                world.state.rng.rand = orig_rand  # type: ignore[assignment]
                world.state.particles._rand = orig_particles_rand
                world.state.sprite_effects._rand = orig_sprite_effects_rand
                projectiles_mod._within_native_find_radius = orig_within
                presentation_step_mod.run_projectile_decal_hooks = orig_run_projectile_decal_hooks

            if not use_outside_draws:
                draws = max(0, int(inter_tick_rand_draws))
                for _ in range(draws):
                    world.state.rng.rand()
    finally:
        world.state.rng.rand = orig_rand  # type: ignore[assignment]
        world.state.particles._rand = orig_particles_rand
        world.state.sprite_effects._rand = orig_sprite_effects_rand
        projectiles_mod._within_native_find_radius = orig_within
        presentation_step_mod.run_projectile_decal_hooks = orig_run_projectile_decal_hooks

    near_misses.sort(key=lambda row: float(row.margin))
    collision_hits.sort(key=lambda row: (int(row.proj_index or -1), int(row.step or -1), int(row.creature_idx or -1)))

    creature_diffs_top = _summarize_creature_diffs(capture_creatures, world)  # ty:ignore[invalid-argument-type]
    projectile_diffs_top = _summarize_projectile_diffs(capture_projectiles, world)  # ty:ignore[invalid-argument-type]
    creature_capture_only, creature_rewrite_only = _collect_creature_presence_diffs(capture_creatures, world)  # ty:ignore[invalid-argument-type]
    projectile_capture_only, projectile_rewrite_only = _collect_projectile_presence_diffs(capture_projectiles, world)  # ty:ignore[invalid-argument-type]
    rng_alignment = _summarize_rng_alignment(
        capture_rng_head=[row for row in capture_rng_head if isinstance(row, dict)],  # ty:ignore[not-iterable]
        capture_rng_calls=int(capture_rng_calls),
        rewrite_rng_values=rng_values,
        rewrite_rng_callsites=rng_values_callsites,
    )
    native_caller_gaps_top = _build_native_caller_gaps(rng_alignment)
    fire_bullets_loop_parity = _build_fire_bullets_loop_parity(rng_alignment)

    return FocusTraceReport(
        tick=int(tick),
        hits=int(focus_hits),
        deaths=int(focus_deaths),
        sfx=int(focus_sfx),
        rand_calls_total=int(sum(rng_callsites.values())),
        rng_callsites_top=list(rng_callsites.most_common(64)),
        rng_callsites_head=list(rng_head),
        collision_hits=collision_hits,
        collision_near_misses=near_misses,
        pre_projectiles=pre_projectiles,
        post_projectiles=post_projectiles,
        capture_projectiles=list(capture_projectiles),  # ty:ignore[invalid-argument-type]
        capture_creatures=list(capture_creatures),  # ty:ignore[invalid-argument-type]
        creature_diffs_top=creature_diffs_top,
        creature_capture_only=creature_capture_only,
        creature_rewrite_only=creature_rewrite_only,
        projectile_diffs_top=projectile_diffs_top,
        projectile_capture_only=projectile_capture_only,
        projectile_rewrite_only=projectile_rewrite_only,
        decal_hook_rows=decal_hook_rows,
        rng_alignment=rng_alignment,
        native_caller_gaps_top=native_caller_gaps_top,
        fire_bullets_loop_parity=fire_bullets_loop_parity,
    )


def _print_report(report: FocusTraceReport, *, top_rng: int, near_miss_limit: int, diff_limit: int) -> None:
    print(f"tick={int(report.tick)} hits={int(report.hits)} deaths={int(report.deaths)} sfx={int(report.sfx)}")
    print(f"rand_calls_total={int(report.rand_calls_total)}")

    print("\nrng_value_alignment:")
    align = report.rng_alignment
    print(
        "  "
        f"capture_calls={int(align.capture_calls)} capture_head_len={int(align.capture_head_len)} "
        f"rewrite_calls={int(align.rewrite_calls)} prefix_match={int(align.value_prefix_match)}"
    )
    if align.first_value_mismatch_index is not None:
        print(
            "  "
            f"first_value_mismatch_idx={int(align.first_value_mismatch_index)} "
            f"capture={int(align.first_value_mismatch_capture or 0)} "
            f"rewrite={int(align.first_value_mismatch_rewrite or 0)}"
        )
    if int(align.missing_native_tail_count) > 0:
        print(f"  missing_native_tail={int(align.missing_native_tail_count)}")
        if align.missing_native_tail_callers_top:
            print("  missing_native_tail_callers_top:")
            for caller_static, count in align.missing_native_tail_callers_top:
                print(f"    {caller_static}: {int(count)}")
        if align.missing_native_tail_inferred_callsites_top:
            print("  missing_native_tail_inferred_callsites_top:")
            for callsite, count in align.missing_native_tail_inferred_callsites_top:
                print(f"    {callsite}: {int(count)}")
        if align.missing_native_tail_preview:
            print("  missing_native_tail_preview:")
            for row in align.missing_native_tail_preview[: max(1, int(diff_limit))]:
                inferred = str(row.inferred_rewrite_callsite) if str(row.inferred_rewrite_callsite) else "<unknown>"
                print(
                    "    "
                    f"idx={int(row.index)} caller={row.capture_caller_static} value={int(row.capture_value)} "
                    f"inferred={inferred}"
                )
            if len(align.missing_native_tail_preview) > int(diff_limit):
                print(f"    ... {len(align.missing_native_tail_preview) - int(diff_limit)} more")

    if report.native_caller_gaps_top:
        print("\nnative_caller_gaps_top:")
        for row in report.native_caller_gaps_top[: max(1, int(diff_limit))]:
            label = str(row.native_label) if str(row.native_label) else "<unmapped>"
            inferred = str(row.inferred_rewrite_callsite) if str(row.inferred_rewrite_callsite) else "<unknown>"
            print(
                "  "
                f"{row.native_caller_static} ({label}) "
                f"capture={int(row.capture_count)} rewrite={int(row.rewrite_count)} "
                f"gap={int(row.gap)} inferred={inferred}"
            )
        if len(report.native_caller_gaps_top) > int(diff_limit):
            print(f"  ... {len(report.native_caller_gaps_top) - int(diff_limit)} more")

    if report.fire_bullets_loop_parity is not None:
        parity = report.fire_bullets_loop_parity
        print("\nfire_bullets_loop_parity:")
        print(
            "  "
            f"seed_iterations capture={int(parity.capture_iterations)} rewrite={int(parity.rewrite_iterations)} "
            f"missing={int(parity.missing_iterations)}"
        )
        print(
            "  "
            f"estimated_missing_hits={float(parity.estimated_missing_hits):.3f} "
            f"(iters_per_hit={int(parity.loop_iterations_per_hit)})"
        )
        print(
            "  "
            f"midrange_rerolls capture={int(parity.capture_midrange_rolls)} "
            f"rewrite={int(parity.rewrite_midrange_rolls)}"
        )
        print(
            "  "
            f"farrange_rerolls capture={int(parity.capture_farrange_rolls)} "
            f"rewrite={int(parity.rewrite_farrange_rolls)}"
        )
        print(
            "  "
            f"pre_freeze_rolls capture={int(parity.capture_pre_freeze_rolls)} "
            f"rewrite={int(parity.rewrite_pre_freeze_rolls)}"
        )

    print("\nrng_callsites_top:")
    for key, count in report.rng_callsites_top[: max(1, int(top_rng))]:
        print(f"  {int(count):4d} {key}")

    print("\ncollision_hits:")
    for row in report.collision_hits[: max(1, int(near_miss_limit))]:
        print(
            f"  proj={row.proj_index} step={row.step} creature={row.creature_idx} "
            f"margin={row.margin:.6f} dist={row.dist:.6f} threshold={row.threshold:.6f}"
        )
    if len(report.collision_hits) > int(near_miss_limit):
        print(f"  ... {len(report.collision_hits) - int(near_miss_limit)} more")

    print("\ncollision_near_misses:")
    for row in report.collision_near_misses[: max(1, int(near_miss_limit))]:
        print(
            f"  proj={row.proj_index} step={row.step} creature={row.creature_idx} "
            f"margin={row.margin:.6f} dist={row.dist:.6f} threshold={row.threshold:.6f}"
        )
    if len(report.collision_near_misses) > int(near_miss_limit):
        print(f"  ... {len(report.collision_near_misses) - int(near_miss_limit)} more")

    print("\ncreature_diffs_top:")
    for row in report.creature_diffs_top[: max(1, int(diff_limit))]:
        print(
            f"  idx={int(row['index']):3d} hp_delta={float(row['hp_delta']):+.6f} "
            f"hitbox_delta={float(row['hitbox_delta']):+.6f} "
            f"x_delta={float(row['x_delta']):+.6f} y_delta={float(row['y_delta']):+.6f} "
            f"active={bool(row['active_capture'])}/{bool(row['active_rewrite'])}"
        )

    print("\ncreature_presence_diffs:")
    if report.creature_capture_only:
        print("  capture_only_indices:")
        for row in report.creature_capture_only[: max(1, int(diff_limit))]:
            print(
                f"    idx={int(row['index']):3d} type={int(row['type_id'])} "
                f"hp={float(row['hp']):.6f} hitbox={float(row['hitbox_size']):.6f} "
                f"pos=({float(row['pos']['x']):.4f},{float(row['pos']['y']):.4f})"
            )
        if len(report.creature_capture_only) > int(diff_limit):
            print(f"    ... {len(report.creature_capture_only) - int(diff_limit)} more")
    if report.creature_rewrite_only:
        print("  rewrite_only_indices:")
        for row in report.creature_rewrite_only[: max(1, int(diff_limit))]:
            print(
                f"    idx={int(row['index']):3d} type={int(row['type_id'])} "
                f"hp={float(row['hp']):.6f} hitbox={float(row['hitbox_size']):.6f} "
                f"pos=({float(row['pos']['x']):.4f},{float(row['pos']['y']):.4f})"
            )
        if len(report.creature_rewrite_only) > int(diff_limit):
            print(f"    ... {len(report.creature_rewrite_only) - int(diff_limit)} more")
    if not report.creature_capture_only and not report.creature_rewrite_only:
        print("  none")

    print("\nprojectile_diffs_top:")
    for row in report.projectile_diffs_top[: max(1, int(diff_limit))]:
        print(
            f"  idx={int(row['index']):3d} life_delta={float(row['life_delta']):+.6f} "
            f"damage_pool_delta={float(row['damage_pool_delta']):+.6f} "
            f"x_delta={float(row['x_delta']):+.6f} y_delta={float(row['y_delta']):+.6f} "
            f"active={bool(row['active_capture'])}/{bool(row['active_rewrite'])}"
        )

    print("\nprojectile_presence_diffs:")
    if report.projectile_capture_only:
        print("  capture_only_indices:")
        for row in report.projectile_capture_only[: max(1, int(diff_limit))]:
            print(
                f"    idx={int(row['index']):3d} type={int(row['type_id'])} "
                f"life={float(row['life_timer']):.6f} damage_pool={float(row['damage_pool']):.6f} "
                f"pos=({float(row['pos']['x']):.4f},{float(row['pos']['y']):.4f})"
            )
        if len(report.projectile_capture_only) > int(diff_limit):
            print(f"    ... {len(report.projectile_capture_only) - int(diff_limit)} more")
    if report.projectile_rewrite_only:
        print("  rewrite_only_indices:")
        for row in report.projectile_rewrite_only[: max(1, int(diff_limit))]:
            print(
                f"    idx={int(row['index']):3d} type={int(row['type_id'])} "
                f"life={float(row['life_timer']):.6f} damage_pool={float(row['damage_pool']):.6f} "
                f"pos=({float(row['pos']['x']):.4f},{float(row['pos']['y']):.4f})"
            )
        if len(report.projectile_rewrite_only) > int(diff_limit):
            print(f"    ... {len(report.projectile_rewrite_only) - int(diff_limit)} more")
    if not report.projectile_capture_only and not report.projectile_rewrite_only:
        print("  none")

    print("\ndecal_hook_rows:")
    if report.decal_hook_rows:
        for row in report.decal_hook_rows[: max(1, int(diff_limit))]:
            print(
                f"  hook={int(row.hook_index):2d} type={int(row.type_id):2d} handled={int(bool(row.handled))} "
                f"rng_draws={int(row.rng_draws):3d} target=({float(row.target_x):.4f},{float(row.target_y):.4f})"
            )
        if len(report.decal_hook_rows) > int(diff_limit):
            print(f"  ... {len(report.decal_hook_rows) - int(diff_limit)} more")
    else:
        print("  none")


def _build_arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Trace a single original-capture tick with rewrite RNG callsites and collision near-miss diagnostics.",
    )
    parser.add_argument("capture", type=Path, help="capture file (.json/.json.gz)")
    parser.add_argument("--tick", type=int, required=True, help="tick index to trace")
    parser.add_argument(
        "--near-miss-threshold",
        type=float,
        default=0.35,
        help="include non-hit collision checks with 0 <= margin <= threshold",
    )
    parser.add_argument("--near-miss-limit", type=int, default=24, help="max rows for hit/near-miss print")
    parser.add_argument("--top-rng", type=int, default=24, help="max RNG callsites to print")
    parser.add_argument("--diff-limit", type=int, default=16, help="max sample-diff rows to print")
    parser.add_argument(
        "--inter-tick-rand-draws",
        type=int,
        default=1,
        help="extra rand draws between ticks (native console loop parity)",
    )
    parser.add_argument(
        "--json-out",
        nargs="?",
        default=None,
        const=_JSON_OUT_AUTO,
        help=(
            "optional JSON output path "
            "(default when flag is present: artifacts/frida/reports/focus_trace_tick<TICK>_latest.json)"
        ),
    )
    return parser


def main(argv: list[str] | None = None) -> int:
    parser = _build_arg_parser()
    args = parser.parse_args(argv)
    json_out_path = _resolve_json_out_path(args.json_out, tick=int(args.tick))

    report = trace_focus_tick(
        capture_path=Path(args.capture),
        tick=int(args.tick),
        near_miss_threshold=max(0.0, float(args.near_miss_threshold)),
        inter_tick_rand_draws=max(0, int(args.inter_tick_rand_draws)),
    )
    _print_report(
        report,
        top_rng=max(1, int(args.top_rng)),
        near_miss_limit=max(1, int(args.near_miss_limit)),
        diff_limit=max(1, int(args.diff_limit)),
    )

    if json_out_path is not None:
        payload = {
            "tick": int(report.tick),
            "hits": int(report.hits),
            "deaths": int(report.deaths),
            "sfx": int(report.sfx),
            "rand_calls_total": int(report.rand_calls_total),
            "rng_callsites_top": [[key, int(count)] for key, count in report.rng_callsites_top],
            "rng_callsites_head": list(report.rng_callsites_head),
            "collision_hits": [asdict(row) for row in report.collision_hits],
            "collision_near_misses": [asdict(row) for row in report.collision_near_misses],
            "pre_projectiles": list(report.pre_projectiles),
            "post_projectiles": list(report.post_projectiles),
            "capture_projectiles": list(report.capture_projectiles),
            "capture_creatures": list(report.capture_creatures),
            "creature_diffs_top": list(report.creature_diffs_top),
            "creature_capture_only": list(report.creature_capture_only),
            "creature_rewrite_only": list(report.creature_rewrite_only),
            "projectile_diffs_top": list(report.projectile_diffs_top),
            "projectile_capture_only": list(report.projectile_capture_only),
            "projectile_rewrite_only": list(report.projectile_rewrite_only),
            "decal_hook_rows": [asdict(row) for row in report.decal_hook_rows],
            "rng_alignment": asdict(report.rng_alignment),
            "native_caller_gaps_top": [asdict(row) for row in report.native_caller_gaps_top],
            "fire_bullets_loop_parity": (
                asdict(report.fire_bullets_loop_parity) if report.fire_bullets_loop_parity is not None else None
            ),
        }
        out_path = json_out_path
        out_path.parent.mkdir(parents=True, exist_ok=True)
        out_path.write_text(json.dumps(payload, indent=2, sort_keys=True), encoding="utf-8")
        print(f"\njson_report={out_path}")

    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations
from dataclasses import dataclass
import struct

from grim.geom import Rect, Vec2
from grim.config import (
    KEYBIND_UNBOUND_CODE,
    default_player_keybind_block,
    hud_indicator_enabled_for_player,
    player_keybind_value,
    set_hud_indicator_for_player,
    set_player_keybind_value,
)

import pyray as rl

from grim.fonts.small import SmallFontData, draw_small_text, load_small_font, measure_small_text_width

from ..menu import (
    MENU_PANEL_HEIGHT,
    MENU_PANEL_WIDTH,
    MenuView,
)
from ...ui.menu_panel import draw_classic_menu_panel
from .base import PANEL_TIMELINE_END_MS, PANEL_TIMELINE_START_MS, PanelMenuView
from .controls_labels import (
    PICK_PERK_BIND_SLOT,
    RELOAD_BIND_SLOT,
    controls_method_values,
    controls_rebind_slot_plan,
    input_configure_for_label,
    input_scheme_label,
)
from ...input_codes import INPUT_CODE_UNBOUND, capture_first_pressed_input_code, input_code_name
from .hit_test import mouse_inside_rect_with_padding
from ..types import GameState


# Measured from ui_render_trace_oracle_1024x768.json (state_3:Configure for:, timeline=300).
CONTROLS_LEFT_PANEL_POS_X = -165.0
CONTROLS_LEFT_PANEL_POS_Y = 200.0
CONTROLS_RIGHT_PANEL_POS_X = 590.0
CONTROLS_RIGHT_PANEL_POS_Y = 110.0
CONTROLS_RIGHT_PANEL_HEIGHT = 378.0
CONTROLS_BACK_POS_X = -155.0
CONTROLS_BACK_POS_Y = 420.0

# `ui_menu_item_update`: idle rebind value tint (rgb 70,180,240 @ alpha 0.6).
CONTROLS_REBIND_VALUE_COLOR = rl.Color(70, 180, 240, 153)
CONTROLS_REBIND_HOVER_COLOR = rl.Color(200, 230, 250, 230)
CONTROLS_REBIND_ACTIVE_COLOR = rl.Color(255, 228, 170, 255)

_AXIS_REBIND_SLOTS = frozenset((9, 10, 11, 12))


@dataclass(frozen=True, slots=True)
class _DropdownLayout:
    pos: Vec2
    width: float
    header_h: float
    row_h: float
    rows_y0: float
    full_h: float
    arrow_pos: Vec2
    arrow_size: Vec2
    text_pos: Vec2
    text_scale: float


@dataclass(frozen=True, slots=True)
class _RebindRowLayout:
    label: str
    slot: int
    row_y: float
    value_pos: Vec2
    value_rect: Rect


class ControlsMenuView(PanelMenuView):
    def __init__(self, state: GameState) -> None:
        super().__init__(
            state,
            title="Controls",
            back_action="open_options",
            panel_pos=Vec2(CONTROLS_LEFT_PANEL_POS_X, CONTROLS_LEFT_PANEL_POS_Y),
            back_pos=Vec2(CONTROLS_BACK_POS_X, CONTROLS_BACK_POS_Y),
        )
        self._small_font: SmallFontData | None = None
        self._text_controls: rl.Texture | None = None
        self._drop_on: rl.Texture | None = None
        self._drop_off: rl.Texture | None = None
        self._check_on: rl.Texture | None = None
        self._check_off: rl.Texture | None = None

        self._config_player = 1
        self._move_method_open = False
        self._aim_method_open = False
        self._player_profile_open = False
        self._dirty = False
        self._rebind_slot: int | None = None
        self._rebind_player_index: int | None = None
        self._rebind_skip_frames = 0

    def open(self) -> None:
        super().open()
        cache = self._ensure_cache()
        # UI elements used by the classic controls screen.
        self._text_controls = cache.get_or_load("ui_textControls", "ui/ui_textControls.jaz").texture
        self._drop_on = cache.get_or_load("ui_dropOn", "ui/ui_dropDownOn.jaz").texture
        self._drop_off = cache.get_or_load("ui_dropOff", "ui/ui_dropDownOff.jaz").texture
        self._check_on = cache.get_or_load("ui_checkOn", "ui/ui_checkOn.jaz").texture
        self._check_off = cache.get_or_load("ui_checkOff", "ui/ui_checkOff.jaz").texture
        self._config_player = max(1, min(4, int(self._config_player)))
        self._move_method_open = False
        self._aim_method_open = False
        self._player_profile_open = False
        self._dirty = False
        self._clear_rebind_capture()

    def update(self, dt: float) -> None:
        super().update(dt)
        if self._closing:
            return
        entry = self._entry
        if entry is None or not self._entry_enabled(entry):
            return
        panel_scale, _local_y_shift = self._menu_item_scale(0)
        left_top_left = self._left_panel_top_left(panel_scale)
        right_top_left = self._right_panel_top_left(panel_scale)
        click_consumed = self._update_method_dropdowns(left_top_left=left_top_left, panel_scale=panel_scale)
        if not click_consumed:
            click_consumed = self._update_rebind_capture(right_top_left=right_top_left, panel_scale=panel_scale)
        if (not click_consumed) and self._update_direction_arrow_checkbox(
            left_top_left=left_top_left,
            panel_scale=panel_scale,
            enabled=self._checkbox_enabled(),
        ):
            self._dirty = True

    def _begin_close_transition(self, action: str) -> None:
        if self._dirty:
            try:
                self._state.config.save()
            except Exception:
                pass
            self._dirty = False
        super()._begin_close_transition(action)

    def _ensure_small_font(self) -> SmallFontData:
        if self._small_font is not None:
            return self._small_font
        missing_assets: list[str] = []
        self._small_font = load_small_font(self._state.assets_dir, missing_assets)
        return self._small_font

    def _current_player_index(self) -> int:
        return max(0, min(3, int(self._config_player) - 1))

    def _rebind_active(self) -> bool:
        return self._rebind_slot is not None and self._rebind_player_index is not None

    def _clear_rebind_capture(self) -> None:
        self._rebind_slot = None
        self._rebind_player_index = None
        self._rebind_skip_frames = 0

    def _start_rebind_capture(self, *, slot: int, player_index: int) -> None:
        self._rebind_slot = int(slot)
        self._rebind_player_index = max(0, min(3, int(player_index)))
        self._move_method_open = False
        self._aim_method_open = False
        self._player_profile_open = False
        # Ignore the click that opened capture so Mouse1 is not rebound accidentally.
        self._rebind_skip_frames = 1

    @staticmethod
    def _slot_is_axis(slot: int) -> bool:
        return int(slot) in _AXIS_REBIND_SLOTS

    @staticmethod
    def _capture_prompt_for_slot(slot: int) -> str:
        if ControlsMenuView._slot_is_axis(int(slot)):
            return "<press axis>"
        return "<press input>"

    def _slot_default_key(self, *, player_index: int, slot: int) -> int:
        slot_idx = int(slot)
        if slot_idx == PICK_PERK_BIND_SLOT:
            return 0x101
        if slot_idx == RELOAD_BIND_SLOT:
            return 0x102
        defaults = default_player_keybind_block(int(player_index))
        if 0 <= slot_idx < len(defaults):
            return int(defaults[slot_idx])
        return int(KEYBIND_UNBOUND_CODE)

    def _slot_key(self, *, player_index: int, slot: int) -> int:
        slot_idx = int(slot)
        if slot_idx == PICK_PERK_BIND_SLOT:
            return int(self._state.config.data.get("keybind_pick_perk", 0x101) or 0x101)
        if slot_idx == RELOAD_BIND_SLOT:
            return int(self._state.config.data.get("keybind_reload", 0x102) or 0x102)
        return int(player_keybind_value(self._state.config.data, player_index=int(player_index), slot_index=slot_idx))

    def _set_slot_key(self, *, player_index: int, slot: int, code: int) -> None:
        slot_idx = int(slot)
        value = int(code)
        if slot_idx == PICK_PERK_BIND_SLOT:
            self._state.config.data["keybind_pick_perk"] = value
            return
        if slot_idx == RELOAD_BIND_SLOT:
            self._state.config.data["keybind_reload"] = value
            return
        set_player_keybind_value(
            self._state.config.data,
            player_index=int(player_index),
            slot_index=slot_idx,
            value=value,
        )

    def _left_panel_top_left(self, panel_scale: float) -> Vec2:
        panel_w = MENU_PANEL_WIDTH * panel_scale
        _, slide_x = MenuView._ui_element_anim(
            self,
            index=1,
            start_ms=PANEL_TIMELINE_START_MS,
            end_ms=PANEL_TIMELINE_END_MS,
            width=panel_w,
        )
        return (
            Vec2(
                self._panel_pos.x + slide_x,
                self._panel_pos.y + self._widescreen_y_shift,
            )
            + self._panel_offset * panel_scale
        )

    def _right_panel_top_left(self, panel_scale: float) -> Vec2:
        panel_w = MENU_PANEL_WIDTH * panel_scale
        _, slide_x = MenuView._ui_element_anim(
            self,
            index=3,
            start_ms=PANEL_TIMELINE_START_MS,
            end_ms=PANEL_TIMELINE_END_MS,
            width=panel_w,
            direction_flag=1,
        )
        return (
            Vec2(
                CONTROLS_RIGHT_PANEL_POS_X + slide_x,
                CONTROLS_RIGHT_PANEL_POS_Y + self._widescreen_y_shift,
            )
            + self._panel_offset * panel_scale
        )

    def _direction_arrow_enabled(self) -> bool:
        return bool(
            hud_indicator_enabled_for_player(self._state.config.data, player_index=int(self._current_player_index()))
        )

    def _set_direction_arrow_enabled(self, enabled: bool) -> None:
        set_hud_indicator_for_player(
            self._state.config.data,
            player_index=int(self._current_player_index()),
            enabled=bool(enabled),
        )

    def _checkbox_enabled(self) -> bool:
        return not (self._move_method_open or self._aim_method_open or self._rebind_active())

    def _checkbox_hovered(self, *, left_top_left: Vec2, panel_scale: float, enabled: bool) -> bool:
        if not enabled:
            return False
        check_on = self._check_on
        check_off = self._check_off
        if check_on is None or check_off is None:
            return False
        font = self._ensure_small_font()
        text_scale = 1.0 * panel_scale
        label = "Show direction arrow"
        check_pos = Vec2(left_top_left.x + 213.0 * panel_scale, left_top_left.y + 174.0 * panel_scale)
        label_w = measure_small_text_width(font, label, text_scale)
        rect_w = float(check_on.width) * panel_scale + 6.0 * panel_scale + label_w
        rect_h = max(float(check_on.height) * panel_scale, font.cell_size * text_scale)
        mouse_pos = Vec2.from_xy(rl.get_mouse_position())
        return Rect.from_top_left(check_pos, rect_w, rect_h).contains(mouse_pos)

    def _update_direction_arrow_checkbox(self, *, left_top_left: Vec2, panel_scale: float, enabled: bool) -> bool:
        if not enabled:
            return False
        check_on = self._check_on
        check_off = self._check_off
        if check_on is None or check_off is None:
            return False

        hovered = self._checkbox_hovered(left_top_left=left_top_left, panel_scale=panel_scale, enabled=enabled)
        if hovered and rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT):
            self._set_direction_arrow_enabled(not self._direction_arrow_enabled())
            return True
        return False

    def _rebind_sections(
        self,
        *,
        player_index: int,
        aim_scheme: int,
        move_mode: int,
    ) -> tuple[tuple[str, tuple[tuple[str, int], ...]], ...]:
        aim_rows, move_rows, misc_rows = controls_rebind_slot_plan(
            aim_scheme=aim_scheme,
            move_mode=move_mode,
            player_index=player_index,
        )
        sections: list[tuple[str, tuple[tuple[str, int], ...]]] = [("Aiming", aim_rows), ("Moving", move_rows)]
        if misc_rows:
            sections.append(("Misc", misc_rows))
        return tuple(sections)

    def _collect_rebind_rows(
        self,
        *,
        right_top_left: Vec2,
        panel_scale: float,
        player_index: int,
        sections: tuple[tuple[str, tuple[tuple[str, int], ...]], ...],
    ) -> tuple[_RebindRowLayout, ...]:
        font = self._ensure_small_font()
        text_scale = 1.0 * panel_scale
        rows: list[_RebindRowLayout] = []
        y = right_top_left.y + 64.0 * panel_scale
        for _section_title, section_rows in sections:
            row_y = y + 18.0 * panel_scale
            for label, slot in section_rows:
                key_code = int(self._slot_key(player_index=player_index, slot=slot))
                value_text = input_code_name(key_code)
                value_pos = Vec2(right_top_left.x + 180.0 * panel_scale, row_y)
                value_w = max(60.0 * panel_scale, measure_small_text_width(font, value_text, text_scale))
                value_rect = Rect.from_top_left(
                    Vec2(value_pos.x - 2.0 * panel_scale, row_y - 2.0 * panel_scale),
                    value_w + 4.0 * panel_scale,
                    14.0 * panel_scale,
                )
                rows.append(
                    _RebindRowLayout(
                        label=str(label),
                        slot=int(slot),
                        row_y=float(row_y),
                        value_pos=value_pos,
                        value_rect=value_rect,
                    )
                )
                row_y += 16.0 * panel_scale
            y = row_y + 8.0 * panel_scale
        return tuple(rows)

    def _update_rebind_capture(self, *, right_top_left: Vec2, panel_scale: float) -> bool:
        player_idx = self._current_player_index()
        aim_scheme, move_mode = controls_method_values(self._state.config.data, player_index=player_idx)
        sections = self._rebind_sections(player_index=player_idx, aim_scheme=aim_scheme, move_mode=move_mode)
        rows = self._collect_rebind_rows(
            right_top_left=right_top_left,
            panel_scale=panel_scale,
            player_index=player_idx,
            sections=sections,
        )

        if self._rebind_active():
            active_slot = int(self._rebind_slot or 0)
            active_player = int(self._rebind_player_index or 0)
            if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE) or rl.is_mouse_button_pressed(
                rl.MouseButton.MOUSE_BUTTON_RIGHT
            ):
                self._clear_rebind_capture()
                return True

            if rl.is_key_pressed(rl.KeyboardKey.KEY_BACKSPACE):
                self._set_slot_key(
                    player_index=active_player,
                    slot=active_slot,
                    code=self._slot_default_key(player_index=active_player, slot=active_slot),
                )
                self._dirty = True
                self._clear_rebind_capture()
                return True

            if rl.is_key_pressed(rl.KeyboardKey.KEY_DELETE):
                self._set_slot_key(player_index=active_player, slot=active_slot, code=INPUT_CODE_UNBOUND)
                self._dirty = True
                self._clear_rebind_capture()
                return True

            if self._rebind_skip_frames > 0:
                self._rebind_skip_frames = max(0, int(self._rebind_skip_frames) - 1)
                return True

            axis_only = self._slot_is_axis(active_slot)
            captured = capture_first_pressed_input_code(
                player_index=active_player,
                include_keyboard=not axis_only,
                include_mouse=not axis_only,
                include_gamepad=not axis_only,
                include_axes=axis_only,
                axis_threshold=0.5,
            )
            if captured is not None:
                self._set_slot_key(player_index=active_player, slot=active_slot, code=int(captured))
                self._dirty = True
                self._clear_rebind_capture()
            return True

        if self._move_method_open or self._aim_method_open or self._player_profile_open:
            return False

        if not rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT):
            return False
        mouse = Vec2.from_xy(rl.get_mouse_position())
        for row in rows:
            if row.value_rect.contains(mouse):
                self._start_rebind_capture(slot=row.slot, player_index=player_idx)
                return True
        return False

    @staticmethod
    def _coerce_blob(raw: object, size: int) -> bytearray:
        values = bytearray(raw) if isinstance(raw, (bytes, bytearray)) else bytearray()
        if len(values) < size:
            values.extend(b"\x00" * (size - len(values)))
        if len(values) > size:
            del values[size:]
        return values

    def _set_player_move_mode(self, *, player_index: int, move_mode: int) -> None:
        config = self._state.config
        raw = self._coerce_blob(config.data.get("unknown_1c"), 0x28)
        idx = max(0, min(3, int(player_index)))
        struct.pack_into("<I", raw, idx * 4, int(move_mode))
        config.data["unknown_1c"] = bytes(raw)

    def _set_player_aim_scheme(self, *, player_index: int, aim_scheme: int) -> None:
        config = self._state.config
        idx = max(0, min(3, int(player_index)))
        scheme = int(aim_scheme)
        if idx == 0:
            config.data["unknown_44"] = scheme
            return
        if idx == 1:
            config.data["unknown_48"] = scheme
            return
        raw = self._coerce_blob(config.data.get("unknown_4c"), 0x20)
        struct.pack_into("<I", raw, (idx - 2) * 4, scheme)
        config.data["unknown_4c"] = bytes(raw)

    @staticmethod
    def _move_method_items(*, move_mode: int) -> tuple[str, ...]:
        items = [input_scheme_label(1), input_scheme_label(2), input_scheme_label(3)]
        if int(move_mode) == 4:
            items.append(input_scheme_label(4))
        return tuple(items)

    def _dropdown_layout(self, *, pos: Vec2, items: tuple[str, ...], scale: float) -> _DropdownLayout:
        font = self._ensure_small_font()
        text_scale = 1.0 * scale
        max_label_w = 0.0
        for label in items:
            max_label_w = max(max_label_w, measure_small_text_width(font, label, text_scale))
        width = max_label_w + 48.0 * scale
        header_h = 16.0 * scale
        row_h = 16.0 * scale
        full_h = (float(len(items)) * 16.0 + 24.0) * scale
        arrow = 16.0 * scale
        return _DropdownLayout(
            pos=pos,
            width=width,
            header_h=header_h,
            row_h=row_h,
            rows_y0=pos.y + 17.0 * scale,
            full_h=full_h,
            arrow_pos=Vec2(pos.x + width - arrow - 1.0 * scale, pos.y),
            arrow_size=Vec2(arrow, arrow),
            text_pos=pos + Vec2(4.0 * scale, 1.0 * scale),
            text_scale=text_scale,
        )

    def _update_dropdown(
        self,
        *,
        layout: _DropdownLayout,
        item_count: int,
        is_open: bool,
        enabled: bool,
        scale: float,
    ) -> tuple[bool, int | None, bool]:
        mouse = rl.get_mouse_position()
        click = rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT)
        hovered_header = bool(enabled) and mouse_inside_rect_with_padding(
            mouse,
            pos=layout.pos,
            width=layout.width,
            height=14.0 * scale,
        )
        if hovered_header and click:
            return (not is_open), None, True
        if not is_open:
            return is_open, None, False

        list_hovered = Rect.from_top_left(layout.pos, layout.width, layout.full_h).contains(Vec2.from_xy(mouse))
        if click and not list_hovered:
            return False, None, True

        for idx in range(item_count):
            item_y = layout.rows_y0 + layout.row_h * float(idx)
            hovered = bool(enabled) and mouse_inside_rect_with_padding(
                mouse,
                pos=Vec2(layout.pos.x, item_y),
                width=layout.width,
                height=14.0 * scale,
            )
            if hovered and click:
                return False, idx, True

        return is_open, None, False

    def _update_method_dropdowns(self, *, left_top_left: Vec2, panel_scale: float) -> bool:
        config = self._state.config
        player_idx = self._current_player_index()
        aim_scheme, move_mode = controls_method_values(config.data, player_index=player_idx)
        move_items = self._move_method_items(move_mode=move_mode)
        aim_items = tuple(input_configure_for_label(i) for i in range(5))
        player_items = ("Player 1", "Player 2", "Player 3", "Player 4")

        move_layout = self._dropdown_layout(
            pos=Vec2(left_top_left.x + 214.0 * panel_scale, left_top_left.y + 144.0 * panel_scale),
            items=move_items,
            scale=panel_scale,
        )
        aim_layout = self._dropdown_layout(
            pos=Vec2(left_top_left.x + 214.0 * panel_scale, left_top_left.y + 102.0 * panel_scale),
            items=aim_items,
            scale=panel_scale,
        )
        player_layout = self._dropdown_layout(
            pos=Vec2(left_top_left.x + 340.0 * panel_scale, left_top_left.y + 56.0 * panel_scale),
            items=player_items,
            scale=panel_scale,
        )

        rebind_active = self._rebind_active()
        move_enabled = not (self._aim_method_open or self._player_profile_open or rebind_active)
        aim_enabled = not (self._move_method_open or self._player_profile_open or rebind_active)
        player_enabled = not (self._move_method_open or self._aim_method_open or rebind_active)

        self._move_method_open, move_selected, consumed = self._update_dropdown(
            layout=move_layout,
            item_count=len(move_items),
            is_open=self._move_method_open,
            enabled=move_enabled,
            scale=panel_scale,
        )
        if move_selected is not None:
            self._set_player_move_mode(player_index=player_idx, move_mode=move_selected + 1)
            self._dirty = True
        if consumed:
            return True

        self._aim_method_open, aim_selected, consumed = self._update_dropdown(
            layout=aim_layout,
            item_count=len(aim_items),
            is_open=self._aim_method_open,
            enabled=aim_enabled,
            scale=panel_scale,
        )
        if aim_selected is not None:
            self._set_player_aim_scheme(player_index=player_idx, aim_scheme=aim_selected)
            self._dirty = True
        if consumed:
            return True

        self._player_profile_open, player_selected, consumed = self._update_dropdown(
            layout=player_layout,
            item_count=len(player_items),
            is_open=self._player_profile_open,
            enabled=player_enabled,
            scale=panel_scale,
        )
        if player_selected is not None:
            self._config_player = max(1, min(4, player_selected + 1))
        if consumed:
            return True

        return False

    def _draw_panel(self) -> None:
        assets = self._assets
        if assets is None or assets.panel is None:
            return
        panel = assets.panel

        fx_detail = bool(self._state.config.data.get("fx_detail_0", 0))
        panel_scale, _local_y_shift = self._menu_item_scale(0)
        panel_w = MENU_PANEL_WIDTH * panel_scale

        # Left (controls options) panel: standard 254px height => a single quad.
        left_top_left = self._left_panel_top_left(panel_scale)
        left_h = MENU_PANEL_HEIGHT * panel_scale
        draw_classic_menu_panel(
            panel,
            dst=rl.Rectangle(left_top_left.x, left_top_left.y, panel_w, left_h),
            tint=rl.WHITE,
            shadow=fx_detail,
        )

        # Right (configured bindings) panel: tall 378px panel rendered as 3 vertical slices.
        right_top_left = self._right_panel_top_left(panel_scale)
        right_h = float(CONTROLS_RIGHT_PANEL_HEIGHT) * panel_scale
        draw_classic_menu_panel(
            panel,
            dst=rl.Rectangle(right_top_left.x, right_top_left.y, panel_w, right_h),
            tint=rl.WHITE,
            shadow=fx_detail,
            # Original ui_element_slot_40 sets direction_flag=1, which mirrors panel UVs.
            flip_x=True,
        )

    def _draw_contents(self) -> None:
        # Positions are expressed relative to the panel top-left corners and scaled with the panel scale.
        panel_scale, _local_y_shift = self._menu_item_scale(0)

        left_top_left = self._left_panel_top_left(panel_scale)
        right_top_left = self._right_panel_top_left(panel_scale)

        font = self._ensure_small_font()
        text_color_full = rl.Color(255, 255, 255, 255)
        text_color_soft = rl.Color(255, 255, 255, 204)
        config = self._state.config
        player_idx = self._current_player_index()
        aim_scheme, move_mode = controls_method_values(config.data, player_index=player_idx)
        move_items = self._move_method_items(move_mode=move_mode)
        aim_items = tuple(input_configure_for_label(i) for i in range(5))
        player_items = ("Player 1", "Player 2", "Player 3", "Player 4")
        move_selected = max(0, min(len(move_items) - 1, int(move_mode) - 1))
        aim_selected = max(0, min(len(aim_items) - 1, int(aim_scheme)))
        player_selected = max(0, min(len(player_items) - 1, player_idx))
        move_layout = self._dropdown_layout(
            pos=Vec2(left_top_left.x + 214.0 * panel_scale, left_top_left.y + 144.0 * panel_scale),
            items=move_items,
            scale=panel_scale,
        )
        aim_layout = self._dropdown_layout(
            pos=Vec2(left_top_left.x + 214.0 * panel_scale, left_top_left.y + 102.0 * panel_scale),
            items=aim_items,
            scale=panel_scale,
        )
        player_layout = self._dropdown_layout(
            pos=Vec2(left_top_left.x + 340.0 * panel_scale, left_top_left.y + 56.0 * panel_scale),
            items=player_items,
            scale=panel_scale,
        )

        # --- Left panel: "Configure for" + method selectors (state_3 in trace) ---
        if self._text_controls is not None:
            MenuView._draw_ui_quad(
                texture=self._text_controls,
                src=rl.Rectangle(0.0, 0.0, float(self._text_controls.width), float(self._text_controls.height)),
                dst=rl.Rectangle(
                    left_top_left.x + 206.0 * panel_scale,
                    left_top_left.y + 44.0 * panel_scale,
                    128.0 * panel_scale,
                    32.0 * panel_scale,
                ),
                origin=rl.Vector2(0.0, 0.0),
                rotation_deg=0.0,
                tint=rl.WHITE,
            )

        draw_small_text(
            font,
            "Configure for:",
            Vec2(left_top_left.x + 339.0 * panel_scale, left_top_left.y + 41.0 * panel_scale),
            1.0 * panel_scale,
            text_color_soft,
        )

        draw_small_text(
            font,
            "Aiming method:",
            Vec2(left_top_left.x + 213.0 * panel_scale, left_top_left.y + 86.0 * panel_scale),
            1.0 * panel_scale,
            text_color_full,
        )

        draw_small_text(
            font,
            "Moving method:",
            Vec2(left_top_left.x + 213.0 * panel_scale, left_top_left.y + 128.0 * panel_scale),
            1.0 * panel_scale,
            text_color_full,
        )

        check_tex = self._check_on if self._direction_arrow_enabled() else self._check_off
        if check_tex is not None:
            MenuView._draw_ui_quad(
                texture=check_tex,
                src=rl.Rectangle(0.0, 0.0, float(check_tex.width), float(check_tex.height)),
                dst=rl.Rectangle(
                    left_top_left.x + 213.0 * panel_scale,
                    left_top_left.y + 174.0 * panel_scale,
                    16.0 * panel_scale,
                    16.0 * panel_scale,
                ),
                origin=rl.Vector2(0.0, 0.0),
                rotation_deg=0.0,
                tint=rl.WHITE,
            )
        checkbox_hovered = self._checkbox_hovered(
            left_top_left=left_top_left,
            panel_scale=panel_scale,
            enabled=self._checkbox_enabled(),
        )
        checkbox_alpha = 255 if checkbox_hovered else 178
        draw_small_text(
            font,
            "Show direction arrow",
            Vec2(left_top_left.x + 235.0 * panel_scale, left_top_left.y + 175.0 * panel_scale),
            1.0 * panel_scale,
            rl.Color(255, 255, 255, checkbox_alpha),
        )

        self._draw_dropdown(
            layout=player_layout,
            items=player_items,
            selected_index=player_selected,
            is_open=self._player_profile_open,
            enabled=not (self._move_method_open or self._aim_method_open or self._rebind_active()),
            scale=panel_scale,
        )
        self._draw_dropdown(
            layout=aim_layout,
            items=aim_items,
            selected_index=aim_selected,
            is_open=self._aim_method_open,
            enabled=not (self._move_method_open or self._player_profile_open or self._rebind_active()),
            scale=panel_scale,
        )
        self._draw_dropdown(
            layout=move_layout,
            items=move_items,
            selected_index=move_selected,
            is_open=self._move_method_open,
            enabled=not (self._aim_method_open or self._player_profile_open or self._rebind_active()),
            scale=panel_scale,
        )

        # --- Right panel: configured bindings list ---
        def _draw_section_heading(title: str, *, y: float) -> None:
            x_heading = right_top_left.x + 44.0 * panel_scale
            draw_small_text(
                font,
                title,
                Vec2(x_heading, y),
                1.0 * panel_scale,
                text_color_full,
            )
            line = rl.Rectangle(
                x_heading,
                y + 13.0 * panel_scale,
                228.0 * panel_scale,
                max(1.0, panel_scale),
            )
            rl.draw_rectangle_lines_ex(line, max(1.0, panel_scale), rl.Color(255, 255, 255, 178))

        draw_small_text(
            font,
            "Configured controls",
            Vec2(right_top_left.x + 120.0 * panel_scale, right_top_left.y + 38.0 * panel_scale),
            1.0 * panel_scale,
            text_color_full,
        )
        header_w = measure_small_text_width(font, "Configured controls", 1.0 * panel_scale)
        header_line = rl.Rectangle(
            right_top_left.x + 120.0 * panel_scale,
            right_top_left.y + 51.0 * panel_scale,
            header_w,
            max(1.0, panel_scale),
        )
        rl.draw_rectangle_lines_ex(header_line, max(1.0, panel_scale), rl.Color(255, 255, 255, 204))

        sections = self._rebind_sections(player_index=player_idx, aim_scheme=aim_scheme, move_mode=move_mode)
        rows = self._collect_rebind_rows(
            right_top_left=right_top_left,
            panel_scale=panel_scale,
            player_index=player_idx,
            sections=sections,
        )
        row_iter = iter(rows)
        mouse = Vec2.from_xy(rl.get_mouse_position())
        dropdown_blocked = self._move_method_open or self._aim_method_open or self._player_profile_open
        rebind_active = self._rebind_active()

        y = right_top_left.y + 64.0 * panel_scale
        for section_title, section_rows in sections:
            _draw_section_heading(section_title, y=y)
            row_y = y + 18.0 * panel_scale
            for _ in section_rows:
                row = next(row_iter)
                label = row.label
                slot = int(row.slot)
                active_row = rebind_active and int(self._rebind_slot or -1) == slot and int(
                    self._rebind_player_index or -1
                ) == player_idx
                hovered_row = (not rebind_active) and (not dropdown_blocked) and row.value_rect.contains(mouse)
                value_text = (
                    self._capture_prompt_for_slot(slot)
                    if active_row
                    else input_code_name(self._slot_key(player_index=player_idx, slot=slot))
                )
                value_pos = row.value_pos

                draw_small_text(
                    font,
                    label,
                    Vec2(right_top_left.x + 52.0 * panel_scale, row_y),
                    1.0 * panel_scale,
                    rl.Color(255, 255, 255, 178),
                )
                value_color = CONTROLS_REBIND_VALUE_COLOR
                if hovered_row:
                    value_color = CONTROLS_REBIND_HOVER_COLOR
                if active_row:
                    value_color = CONTROLS_REBIND_ACTIVE_COLOR
                draw_small_text(
                    font,
                    value_text,
                    value_pos,
                    1.0 * panel_scale,
                    value_color,
                )
                value_w = measure_small_text_width(font, value_text, 1.0 * panel_scale)
                underline_y = row.row_y + 13.0 * panel_scale
                rl.draw_line(
                    int(value_pos.x),
                    int(underline_y),
                    int(value_pos.x + value_w),
                    int(underline_y),
                    value_color,
                )
                row_y += 16.0 * panel_scale
            y = row_y + 8.0 * panel_scale

        if rebind_active and int(self._rebind_player_index or -1) == player_idx:
            hint_pos = Vec2(
                right_top_left.x + 48.0 * panel_scale,
                right_top_left.y + (CONTROLS_RIGHT_PANEL_HEIGHT - 26.0) * panel_scale,
            )
            draw_small_text(
                font,
                "Esc/Right: cancel  Backspace: default  Delete: unbind",
                hint_pos,
                0.85 * panel_scale,
                rl.Color(255, 226, 188, 220),
            )

    def _draw_dropdown(
        self,
        *,
        layout: _DropdownLayout,
        items: tuple[str, ...],
        selected_index: int,
        is_open: bool,
        enabled: bool,
        scale: float,
    ) -> None:
        mouse = rl.get_mouse_position()
        hovered_header = bool(enabled) and mouse_inside_rect_with_padding(
            mouse,
            pos=layout.pos,
            width=layout.width,
            height=14.0 * scale,
        )
        widget_h = layout.full_h if is_open else layout.header_h
        rl.draw_rectangle(int(layout.pos.x), int(layout.pos.y), int(layout.width), int(widget_h), rl.WHITE)
        inner_w = max(0, int(layout.width) - 2)
        inner_h = max(0, int(widget_h) - 2)
        rl.draw_rectangle(int(layout.pos.x) + 1, int(layout.pos.y) + 1, inner_w, inner_h, rl.BLACK)

        if (is_open or hovered_header) and enabled:
            line_h = max(1, int(1.0 * scale))
            rl.draw_rectangle(
                int(layout.pos.x),
                int(layout.pos.y + 15.0 * scale),
                int(layout.width),
                line_h,
                rl.Color(255, 255, 255, 128),
            )

        arrow_tex = self._drop_on if ((is_open or hovered_header) and enabled) else self._drop_off
        if arrow_tex is None:
            arrow_tex = self._drop_off
        if arrow_tex is not None:
            rl.draw_texture_pro(
                arrow_tex,
                rl.Rectangle(0.0, 0.0, float(arrow_tex.width), float(arrow_tex.height)),
                rl.Rectangle(layout.arrow_pos.x, layout.arrow_pos.y, layout.arrow_size.x, layout.arrow_size.y),
                rl.Vector2(0.0, 0.0),
                0.0,
                rl.WHITE,
            )

        idx = max(0, min(len(items) - 1, int(selected_index))) if items else 0
        header_alpha = 242 if ((is_open or hovered_header) and enabled) else 191
        if items:
            draw_small_text(
                self._ensure_small_font(),
                items[idx],
                layout.text_pos,
                layout.text_scale,
                rl.Color(255, 255, 255, header_alpha),
            )

        if not is_open:
            return

        for idx, item in enumerate(items):
            item_y = layout.rows_y0 + layout.row_h * float(idx)
            hovered = bool(enabled) and mouse_inside_rect_with_padding(
                mouse,
                pos=Vec2(layout.pos.x, item_y),
                width=layout.width,
                height=14.0 * scale,
            )
            alpha = 153
            if hovered:
                alpha = 242
            if idx == selected_index:
                alpha = max(alpha, 245)
            draw_small_text(
                self._ensure_small_font(),
                item,
                Vec2(layout.text_pos.x, item_y),
                layout.text_scale,
                rl.Color(255, 255, 255, alpha),
            )



from __future__ import annotations

"""Camera helpers recovered from the original crimsonland.exe.

This module currently models the `camera_update` screen shake logic, which is
global state in the original game.
"""

from grim.geom import Vec2

from .gameplay import GameplayState


def camera_shake_start(state: GameplayState, *, pulses: int, timer: float) -> None:
    """Start a camera shake sequence.

    Mirrors the nuke path in `bonus_apply`, which sets:
      - `camera_shake_pulses = 0x14`
      - `camera_shake_timer = 0.2`
    """

    state.camera_shake_pulses = int(pulses)
    state.camera_shake_timer = float(timer)


def camera_shake_update(state: GameplayState, dt: float) -> None:
    """Update camera shake offsets and timers.

    Port of `camera_update` (crimsonland.exe @ 0x00409500):
      - timer decays at `dt * 3.0`
      - when timer drops below 0, a "pulse" happens:
        - pulses--
        - timer resets to 0.1 (or 0.06 when time scaling is active)
        - offsets jump to new RNG-derived values
    """

    if state.camera_shake_timer <= 0.0:
        state.camera_shake_offset = Vec2()
        return

    state.camera_shake_timer -= float(dt) * 3.0
    if state.camera_shake_timer >= 0.0:
        return

    state.camera_shake_pulses -= 1
    if state.camera_shake_pulses < 1:
        state.camera_shake_timer = 0.0
        return

    time_scale_active = state.bonuses.reflex_boost > 0.0
    state.camera_shake_timer = 0.06 if time_scale_active else 0.1

    # Decompiled logic:
    #   iVar4 = camera_shake_pulses * 0x3c;
    #   iVar1 = rand() % (iVar4 / 0x14) + rand() % 10;
    # ... where (pulses * 0x3c) / 0x14 == pulses * 3.
    max_amp = int(state.camera_shake_pulses) * 3
    if max_amp <= 0:
        state.camera_shake_offset = Vec2()
        state.camera_shake_timer = 0.0
        state.camera_shake_pulses = 0
        return

    rand = state.rng.rand

    mag_x = (int(rand()) % max_amp) + (int(rand()) % 10)
    if (int(rand()) & 1) == 0:
        mag_x = -mag_x
    offset_x = float(mag_x)

    mag_y = (int(rand()) % max_amp) + (int(rand()) % 10)
    if (int(rand()) & 1) == 0:
        mag_y = -mag_y
    offset_y = float(mag_y)
    state.camera_shake_offset = Vec2(offset_x, offset_y)



from __future__ import annotations
from dataclasses import dataclass

from grim.geom import Vec2

from pathlib import Path

import pyray as rl

from grim.fonts.small import SmallFontData, draw_small_text, load_small_font

from ..menu import (
    MENU_PANEL_WIDTH,
    MenuView,
)
from .base import PANEL_TIMELINE_END_MS, PANEL_TIMELINE_START_MS, PanelMenuView

from ..types import GameState


@dataclass(frozen=True, slots=True)
class _ModsContentLayout:
    scale: float
    base_pos: Vec2
    label_pos: Vec2


class ModsMenuView(PanelMenuView):
    def __init__(self, state: GameState) -> None:
        super().__init__(state, title="Mods")
        self._small_font: SmallFontData | None = None
        self._lines: list[str] = []

    def open(self) -> None:
        super().open()
        self._lines = self._build_lines()

    def _ensure_small_font(self) -> SmallFontData:
        if self._small_font is not None:
            return self._small_font
        missing_assets: list[str] = []
        self._small_font = load_small_font(self._state.assets_dir, missing_assets)
        return self._small_font

    def _content_layout(self) -> _ModsContentLayout:
        panel_scale, _local_shift = self._menu_item_scale(0)
        panel_w = MENU_PANEL_WIDTH * panel_scale
        _angle_rad, slide_x = MenuView._ui_element_anim(
            self,
            index=1,
            start_ms=PANEL_TIMELINE_START_MS,
            end_ms=PANEL_TIMELINE_END_MS,
            width=panel_w,
        )
        panel_top_left = (
            Vec2(
                self._panel_pos.x + slide_x,
                self._panel_pos.y + self._widescreen_y_shift,
            )
            + self._panel_offset * panel_scale
        )
        base_pos = panel_top_left + Vec2(212.0 * panel_scale, 32.0 * panel_scale)
        label_pos = base_pos.offset(dx=8.0 * panel_scale)
        return _ModsContentLayout(scale=panel_scale, base_pos=base_pos, label_pos=label_pos)

    def _build_lines(self) -> list[str]:
        mods_dir = self._state.base_dir / "mods"
        dlls: list[Path] = []
        try:
            dlls = sorted(mods_dir.glob("*.dll"))
        except Exception:
            dlls = []

        if not dlls:
            return [
                "No mod DLLs found.",
                "",
                "Expected location:",
                f"  {mods_dir}",
                "",
                "Mod loading is not implemented yet.",
            ]

        lines = [f"Found {len(dlls)} mod DLL(s):", ""]
        for path in dlls[:10]:
            lines.append(f"  {path.name}")
        if len(dlls) > 10:
            lines.append(f"  ... ({len(dlls) - 10} more)")
        lines.append("")
        lines.append("Mod loading is not implemented yet.")
        return lines

    def _draw_contents(self) -> None:
        layout = self._content_layout()
        base_pos = layout.base_pos
        label_pos = layout.label_pos
        scale = layout.scale

        font = self._ensure_small_font()
        title_scale = 1.2 * scale
        text_scale = 1.0 * scale

        title_color = rl.Color(255, 255, 255, 255)
        text_color = rl.Color(255, 255, 255, int(255 * 0.8))

        draw_small_text(font, "MODS", base_pos, title_scale, title_color)
        line_pos = label_pos.offset(dy=44.0 * scale)
        line_step = (font.cell_size + 4.0) * scale
        for line in self._lines:
            draw_small_text(font, line, line_pos, text_scale, text_color)
            line_pos = line_pos.offset(dy=line_step)



from __future__ import annotations

from dataclasses import dataclass

import pyray as rl

from grim.audio import update_audio
from grim.fonts.small import SmallFontData, draw_small_text, load_small_font, measure_small_text_width
from grim.geom import Rect, Vec2

from ...debug import debug_enabled
from ...ui.perk_menu import UiButtonState, UiButtonTextureSet, button_draw, button_update, button_width

from ..menu import (
    MENU_LABEL_ROW_HEIGHT,
    MENU_LABEL_ROW_PLAY_GAME,
    MENU_PANEL_OFFSET_Y,
    MENU_PANEL_WIDTH,
    MenuView,
)
from .base import PANEL_TIMELINE_END_MS, PANEL_TIMELINE_START_MS, PanelMenuView
from .hit_test import mouse_inside_rect_with_padding

from ..types import GameState


@dataclass(slots=True)
class _PlayGameModeEntry:
    key: str
    label: str
    tooltip: str
    action: str
    game_mode: int | None = None
    show_count: bool = False


@dataclass(frozen=True, slots=True)
class _PlayGameContentLayout:
    scale: float
    base_pos: Vec2
    drop_pos: Vec2


@dataclass(frozen=True, slots=True)
class _PlayerCountWidgetLayout:
    pos: Vec2
    width: float
    header_h: float
    row_h: float
    rows_y0: float
    full_h: float
    arrow_pos: Vec2
    arrow_size: Vec2
    text_pos: Vec2
    text_scale: float


class PlayGameMenuView(PanelMenuView):
    """Play Game mode select panel.

    Layout and gating are based on `sub_44ed80` (crimsonland.exe).
    """

    _PLAYER_COUNT_LABELS = ("1 player", "2 players", "3 players", "4 players")

    def __init__(self, state: GameState) -> None:
        super().__init__(
            state,
            title="Play Game",
            panel_offset=Vec2(-63.0, MENU_PANEL_OFFSET_Y),
            panel_height=278.0,
            back_pos=Vec2(-55.0, 462.0),
        )
        self._small_font: SmallFontData | None = None
        self._button_sm: rl.Texture | None = None
        self._button_md: rl.Texture | None = None
        self._button_textures: UiButtonTextureSet | None = None
        self._drop_on: rl.Texture | None = None
        self._drop_off: rl.Texture | None = None

        self._player_list_open = False
        self._dirty = False

        # Hover fade timers for tooltips (0..1000ms-ish; original uses ~0.0009 alpha scale).
        self._tooltip_ms: dict[str, int] = {}
        self._mode_buttons: dict[str, UiButtonState] = {}

    def open(self) -> None:
        super().open()
        cache = self._ensure_cache()
        self._button_sm = cache.get_or_load("ui_buttonSm", "ui/ui_button_64x32.jaz").texture
        self._button_md = cache.get_or_load("ui_buttonMd", "ui/ui_button_128x32.jaz").texture
        self._button_textures = UiButtonTextureSet(button_sm=self._button_sm, button_md=self._button_md)
        self._drop_on = cache.get_or_load("ui_dropOn", "ui/ui_dropDownOn.jaz").texture
        self._drop_off = cache.get_or_load("ui_dropOff", "ui/ui_dropDownOff.jaz").texture
        self._player_list_open = False
        self._dirty = False
        self._tooltip_ms.clear()
        self._mode_buttons.clear()

    def update(self, dt: float) -> None:
        if self._state.audio is not None:
            update_audio(self._state.audio, dt)
        if self._ground is not None:
            self._ground.process_pending()
        self._cursor_pulse_time += min(dt, 0.1) * 1.1
        dt_ms = int(min(dt, 0.1) * 1000.0)

        # Close transition (matches PanelMenuView).
        if self._closing:
            if dt_ms > 0 and self._pending_action is None:
                self._timeline_ms -= dt_ms
                if self._timeline_ms < 0 and self._close_action is not None:
                    self._pending_action = self._close_action
                    self._close_action = None
            return

        if dt_ms > 0:
            self._timeline_ms = min(self._timeline_max_ms, self._timeline_ms + dt_ms)
            if self._timeline_ms >= self._timeline_max_ms:
                self._state.menu_sign_locked = True

        entry = self._entry
        if entry is None:
            return

        enabled = self._entry_enabled(entry)
        hovered_back = enabled and self._hovered_entry(entry)
        self._hovered = hovered_back

        # ESC always goes back; Enter should not auto-back on this screen.
        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE) and enabled:
            self._begin_close_transition(self._back_action)
        if enabled and hovered_back and rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT):
            self._begin_close_transition(self._back_action)

        if hovered_back:
            entry.hover_amount += dt_ms * 6
        else:
            entry.hover_amount -= dt_ms * 2
        entry.hover_amount = max(0, min(1000, entry.hover_amount))

        if entry.ready_timer_ms < 0x100:
            entry.ready_timer_ms = min(0x100, entry.ready_timer_ms + dt_ms)

        if not enabled:
            return

        layout = self._content_layout()
        scale = layout.scale
        base_pos = layout.base_pos

        consumed_click = self._update_player_count(layout.drop_pos, scale)
        if consumed_click:
            return

        mouse = rl.get_mouse_position()
        click = rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT)
        button_enabled = not self._player_list_open

        y = base_pos.y
        entries, y_step, y_start, y_end = self._mode_entries()
        y += y_start * scale
        for mode in entries:
            clicked, hovered = self._update_mode_button(
                mode,
                Vec2(base_pos.x, y),
                scale,
                dt_ms=dt_ms,
                mouse=mouse,
                click=click,
                enabled=button_enabled,
            )
            self._update_tooltip_timer(mode.key, hovered, dt_ms)
            if clicked:
                self._activate_mode(mode)
                return
            y += y_step * scale

        # Decay timers for modes that aren't visible right now.
        visible = {m.key for m in entries}
        for key in list(self._tooltip_ms):
            if key in visible:
                continue
            self._tooltip_ms[key] = max(0, self._tooltip_ms[key] - dt_ms * 2)

    def _begin_close_transition(self, action: str) -> None:
        if self._dirty:
            try:
                self._state.config.save()
            except Exception:
                pass
            self._dirty = False
        super()._begin_close_transition(action)

    def _ensure_small_font(self) -> SmallFontData:
        if self._small_font is not None:
            return self._small_font
        missing_assets: list[str] = []
        self._small_font = load_small_font(self._state.assets_dir, missing_assets)
        return self._small_font

    def _content_layout(self) -> _PlayGameContentLayout:
        panel_scale, _local_shift = self._menu_item_scale(0)
        panel_w = MENU_PANEL_WIDTH * panel_scale
        _angle_rad, slide_x = MenuView._ui_element_anim(
            self,
            index=1,
            start_ms=PANEL_TIMELINE_START_MS,
            end_ms=PANEL_TIMELINE_END_MS,
            width=panel_w,
        )
        panel_top_left = (
            Vec2(
                self._panel_pos.x + slide_x,
                self._panel_pos.y + self._widescreen_y_shift,
            )
            + self._panel_offset * panel_scale
        )

        # `sub_44ed80`:
        #   xy = panel_offset_x + panel_x + 330 - 64  (+ animated X offset)
        #   var_1c = panel_offset_y + panel_y + 50
        base_pos = panel_top_left + Vec2(266.0 * panel_scale, 50.0 * panel_scale)
        drop_pos = base_pos + Vec2(80.0 * panel_scale, 1.0 * panel_scale)

        return _PlayGameContentLayout(
            scale=panel_scale,
            base_pos=base_pos,
            drop_pos=drop_pos,
        )

    def _quests_total_played(self) -> int:
        counts = self._state.status.data.get("quest_play_counts", [])
        if not isinstance(counts, list) or not counts:
            return 0
        # `sub_44ed80` sums 40 ints from game_status_blob+0x104..0x1a4.
        # Our `quest_play_counts` array starts at blob+0xd8, so this is indices 11..50.
        return int(sum(int(v) for v in counts[11:51]))

    def _mode_entries(self) -> tuple[list[_PlayGameModeEntry], float, float, float]:
        config = self._state.config
        status = self._state.status

        # Clamp to a valid range; older configs in the repo can contain 0 here,
        # which would incorrectly hide the Tutorial entry (it is gated on == 1).
        player_count = int(config.data.get("player_count", 1))
        if player_count < 1:
            player_count = 1
        if player_count > len(self._PLAYER_COUNT_LABELS):
            player_count = len(self._PLAYER_COUNT_LABELS)
        quest_unlock = int(status.quest_unlock_index)
        full_version = not self._state.demo_enabled

        quests_total = self._quests_total_played()
        rush_total = int(status.mode_play_count("rush"))
        survival_total = int(status.mode_play_count("survival"))
        # Matches the tutorial placement gating in `sub_44ed80` (excludes Typ-o).
        main_total = quests_total + rush_total + survival_total

        # `sub_44ed80` uses tighter spacing when quest_unlock>=40 and player_count==1.
        tight_spacing = not (quest_unlock < 0x28 or player_count > 1)
        y_step = 28.0 if tight_spacing else 32.0
        y_start = 26.0 if tight_spacing else 32.0

        has_typo = tight_spacing and full_version and player_count == 1
        show_tutorial = player_count == 1

        entries: list[_PlayGameModeEntry] = []
        if show_tutorial and main_total <= 0:
            entries.append(
                _PlayGameModeEntry(
                    key="tutorial",
                    label="Tutorial",
                    tooltip="Learn how to play Crimsonland.",
                    action="start_tutorial",
                    game_mode=8,
                )
            )

        entries.extend(
            [
                _PlayGameModeEntry(
                    key="quests",
                    label=" Quests ",
                    tooltip="Unlock new weapons and perks in Quest mode.",
                    action="open_quests",
                    show_count=True,
                ),
                _PlayGameModeEntry(
                    key="rush",
                    label="  Rush  ",
                    tooltip="Face a rush of aliens in Rush mode.",
                    action="start_rush",
                    game_mode=2,
                    show_count=True,
                ),
                _PlayGameModeEntry(
                    key="survival",
                    label="Survival",
                    tooltip="Gain perks and weapons and fight back.",
                    action="start_survival",
                    game_mode=1,
                    show_count=True,
                ),
            ]
        )

        if has_typo:
            entries.append(
                _PlayGameModeEntry(
                    key="typo",
                    label="Typ'o'Shooter",
                    tooltip="Use your typing skills as the weapon to lay\nthem down.",
                    action="start_typo",
                    game_mode=4,
                    show_count=True,
                )
            )

        if show_tutorial and main_total > 0:
            entries.append(
                _PlayGameModeEntry(
                    key="tutorial",
                    label="Tutorial",
                    tooltip="Learn how to play Crimsonland.",
                    action="start_tutorial",
                    game_mode=8,
                )
            )

        # The y after the last row is used as a tooltip anchor in `sub_44ed80`.
        y_end = y_start + y_step * float(len(entries))
        return entries, y_step, y_start, y_end

    def _mode_button_state(self, mode: _PlayGameModeEntry) -> UiButtonState:
        state = self._mode_buttons.get(mode.key)
        if state is None:
            state = UiButtonState(mode.label)
            self._mode_buttons[mode.key] = state
        else:
            state.label = mode.label
        return state

    def _update_mode_button(
        self,
        mode: _PlayGameModeEntry,
        pos: Vec2,
        scale: float,
        *,
        dt_ms: int,
        mouse: rl.Vector2,
        click: bool,
        enabled: bool,
    ) -> tuple[bool, bool]:
        state = self._mode_button_state(mode)
        state.enabled = bool(enabled)
        font = self._ensure_small_font()
        width = button_width(font, state.label, scale=scale, force_wide=state.force_wide)
        clicked = button_update(state, pos=pos, width=width, dt_ms=float(dt_ms), mouse=mouse, click=bool(click))
        return clicked, state.hovered

    def _activate_mode(self, mode: _PlayGameModeEntry) -> None:
        if mode.game_mode is not None:
            self._state.config.data["game_mode"] = int(mode.game_mode)
            self._dirty = True
        self._begin_close_transition(mode.action)

    def _update_tooltip_timer(self, key: str, hovered: bool, dt_ms: int) -> None:
        value = int(self._tooltip_ms.get(key, 0))
        if hovered:
            value += dt_ms * 6
        else:
            value -= dt_ms * 2
        self._tooltip_ms[key] = max(0, min(1000, value))

    def _player_count_widget_layout(self, pos: Vec2, scale: float) -> _PlayerCountWidgetLayout:
        """Return Play Game player-count dropdown metrics.

        `ui_list_widget_update` (0x43efc0):
          - width = max(label_w) + 0x30
          - header height = 16
          - open height = (count * 16) + 0x18
          - arrow icon = 16x16 at (x + width - 16 - 1, y)
          - selected label at (x + 4, y + 1)
          - list rows start at y + 17, step 16
        """
        font = self._ensure_small_font()
        text_scale = 1.0 * scale
        max_label_w = 0.0
        for label in self._PLAYER_COUNT_LABELS:
            max_label_w = max(max_label_w, measure_small_text_width(font, label, text_scale))
        width = max_label_w + 48.0 * scale
        header_h = 16.0 * scale
        row_h = 16.0 * scale
        full_h = (float(len(self._PLAYER_COUNT_LABELS)) * 16.0 + 24.0) * scale
        arrow = 16.0 * scale
        return _PlayerCountWidgetLayout(
            pos=pos,
            width=width,
            header_h=header_h,
            row_h=row_h,
            rows_y0=pos.y + 17.0 * scale,
            full_h=full_h,
            arrow_pos=Vec2(pos.x + width - arrow - 1.0 * scale, pos.y),
            arrow_size=Vec2(arrow, arrow),
            text_pos=pos + Vec2(4.0 * scale, 1.0 * scale),
            text_scale=text_scale,
        )

    def _update_player_count(self, pos: Vec2, scale: float) -> bool:
        config = self._state.config
        layout = self._player_count_widget_layout(pos, scale)

        mouse = rl.get_mouse_position()
        hovered_header = mouse_inside_rect_with_padding(
            mouse,
            pos=layout.pos,
            width=layout.width,
            height=14.0 * scale,
        )
        if hovered_header and rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT):
            self._player_list_open = not self._player_list_open
            return True

        if not self._player_list_open:
            return False

        # Close if we click outside the dropdown + list.
        list_hovered = Rect.from_top_left(layout.pos, layout.width, layout.full_h).contains(mouse)
        if rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT) and not list_hovered:
            self._player_list_open = False
            return True

        for idx, label in enumerate(self._PLAYER_COUNT_LABELS):
            del label
            item_y = layout.rows_y0 + layout.row_h * float(idx)
            item_hovered = mouse_inside_rect_with_padding(
                mouse,
                pos=Vec2(layout.pos.x, item_y),
                width=layout.width,
                height=14.0 * scale,
            )
            if item_hovered and rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT):
                config.data["player_count"] = idx + 1
                self._dirty = True
                self._player_list_open = False
                return True
        return False

    def _draw_contents(self) -> None:
        assets = self._assets
        if assets is None:
            return
        labels_tex = assets.labels
        layout = self._content_layout()
        base_pos = layout.base_pos
        scale = layout.scale

        font = self._ensure_small_font()
        text_scale = 1.0 * scale
        text_color = rl.Color(255, 255, 255, int(255 * 0.8))

        # `sub_44ed80`: title label at (xy - 64, var_1c - 8), size 128x32.
        title_w = 128.0
        title_h = MENU_LABEL_ROW_HEIGHT
        title_pos = base_pos + Vec2(-64.0 * scale, -8.0 * scale)

        if labels_tex is not None:
            src = rl.Rectangle(
                0.0,
                float(MENU_LABEL_ROW_PLAY_GAME) * MENU_LABEL_ROW_HEIGHT,
                title_w,
                title_h,
            )
            dst = rl.Rectangle(
                title_pos.x,
                title_pos.y,
                title_w * scale,
                title_h * scale,
            )
            MenuView._draw_ui_quad(
                texture=labels_tex,
                src=src,
                dst=dst,
                origin=rl.Vector2(0.0, 0.0),
                rotation_deg=0.0,
                tint=rl.WHITE,
            )
        else:
            rl.draw_text(self._title, int(title_pos.x), int(title_pos.y), int(24 * scale), rl.WHITE)

        entries, y_step, y_start, y_end = self._mode_entries()
        y = base_pos.y + y_start * scale
        show_counts = debug_enabled() and rl.is_key_down(rl.KeyboardKey.KEY_F1)

        if show_counts:
            draw_small_text(
                font,
                "times played:",
                base_pos + Vec2(132.0 * scale, 16.0 * scale),
                text_scale,
                text_color,
            )

        for mode in entries:
            self._draw_mode_button(mode, Vec2(base_pos.x, y), scale)
            if show_counts and mode.show_count:
                self._draw_mode_count(
                    mode.key, Vec2(base_pos.x + 158.0 * scale, y + 8.0 * scale), text_scale, text_color
                )
            y += y_step * scale

        # `sub_44ed80`: the list widget is drawn before tooltips, so tooltips can overlay it.
        self._draw_player_count(layout.drop_pos, scale)
        self._draw_tooltips(entries, base_pos, y_end, scale)

    def _draw_player_count(self, pos: Vec2, scale: float) -> None:
        drop_on = self._drop_on
        drop_off = self._drop_off
        font = self._ensure_small_font()
        layout = self._player_count_widget_layout(pos, scale)

        # `ui_list_widget_update` draws a single bordered black rect for the widget.
        widget_h = layout.full_h if self._player_list_open else layout.header_h
        rl.draw_rectangle(int(layout.pos.x), int(layout.pos.y), int(layout.width), int(widget_h), rl.WHITE)
        inner_w = max(0, int(layout.width) - 2)
        inner_h = max(0, int(widget_h) - 2)
        rl.draw_rectangle(int(layout.pos.x) + 1, int(layout.pos.y) + 1, inner_w, inner_h, rl.BLACK)

        # Arrow icon (the ui_drop* assets are 16x16 icons, not the background).
        mouse = rl.get_mouse_position()
        hovered_header = mouse_inside_rect_with_padding(
            mouse,
            pos=layout.pos,
            width=layout.width,
            height=14.0 * scale,
        )
        arrow_tex = drop_on if (self._player_list_open or hovered_header) else drop_off
        if self._player_list_open or hovered_header:
            line_h = max(1, int(1.0 * scale))
            rl.draw_rectangle(
                int(layout.pos.x),
                int(layout.pos.y + 15.0 * scale),
                int(layout.width),
                line_h,
                rl.Color(255, 255, 255, 128),
            )
        if arrow_tex is not None:
            rl.draw_texture_pro(
                arrow_tex,
                rl.Rectangle(0.0, 0.0, float(arrow_tex.width), float(arrow_tex.height)),
                rl.Rectangle(layout.arrow_pos.x, layout.arrow_pos.y, layout.arrow_size.x, layout.arrow_size.y),
                rl.Vector2(0.0, 0.0),
                0.0,
                rl.WHITE,
            )

        player_count = int(self._state.config.data.get("player_count", 1))
        if player_count < 1:
            player_count = 1
        if player_count > len(self._PLAYER_COUNT_LABELS):
            player_count = len(self._PLAYER_COUNT_LABELS)
        label = self._PLAYER_COUNT_LABELS[player_count - 1]
        header_alpha = 242 if hovered_header else 191  # 0x3f733333 / 0x3f400000
        draw_small_text(font, label, layout.text_pos, layout.text_scale, rl.Color(255, 255, 255, header_alpha))

        if not self._player_list_open:
            return

        for idx, item in enumerate(self._PLAYER_COUNT_LABELS):
            item_y = layout.rows_y0 + layout.row_h * float(idx)
            hovered = mouse_inside_rect_with_padding(
                mouse,
                pos=Vec2(layout.pos.x, item_y),
                width=layout.width,
                height=14.0 * scale,
            )
            alpha = 153  # 0x3f19999a
            if hovered:
                alpha = 242  # 0x3f733333
            if idx == (player_count - 1):
                alpha = max(alpha, 245)  # 0x3f75c28f
            draw_small_text(
                font, item, Vec2(layout.text_pos.x, item_y), layout.text_scale, rl.Color(255, 255, 255, alpha)
            )

    def _draw_mode_button(self, mode: _PlayGameModeEntry, pos: Vec2, scale: float) -> None:
        textures = self._button_textures
        if textures is None:
            return
        if textures.button_sm is None and textures.button_md is None:
            return
        font = self._ensure_small_font()
        state = self._mode_button_state(mode)
        width = button_width(font, state.label, scale=scale, force_wide=state.force_wide)
        button_draw(textures, font, state, pos=pos, width=width, scale=scale)

    def _draw_mode_count(self, key: str, pos: Vec2, scale: float, color: rl.Color) -> None:
        status = self._state.status
        if key == "quests":
            count = self._quests_total_played()
        elif key == "rush":
            count = int(status.mode_play_count("rush"))
        elif key == "survival":
            count = int(status.mode_play_count("survival"))
        elif key == "typo":
            count = int(status.mode_play_count("typo"))
        else:
            return
        draw_small_text(self._ensure_small_font(), f"{count}", pos, scale, color)

    def _draw_tooltips(self, entries: list[_PlayGameModeEntry], base_pos: Vec2, y_end: float, scale: float) -> None:
        # `sub_44ed80` draws these below the mode list based on per-button hover timers.
        font = self._ensure_small_font()
        tooltip_x = base_pos.x - 55.0 * scale
        tooltip_y = base_pos.y + (y_end + 16.0) * scale

        offsets = {
            "quests": (-8.0, 0.0),
            "rush": (32.0, 0.0),
            "survival": (20.0, 0.0),
            "typo": (0.0, -12.0),
            "tutorial": (38.0, 0.0),
        }

        for mode in entries:
            ms = int(self._tooltip_ms.get(mode.key, 0))
            if ms <= 0:
                continue
            alpha_f = min(1.0, float(ms) * 0.0009)
            alpha = int(255 * alpha_f)
            off_x, off_y = offsets.get(mode.key, (0.0, 0.0))
            x = tooltip_x + off_x * scale
            y = tooltip_y + off_y * scale
            for line in mode.tooltip.splitlines():
                draw_small_text(font, line, Vec2(x, y), 1.0 * scale, rl.Color(255, 255, 255, alpha))
                y += font.cell_size * 1.0 * scale



from __future__ import annotations

"""Ordered perk hook registries for deterministic world-step dispatch."""

from .final_revenge import apply_final_revenge_on_player_death
from .reflex_boosted import apply_reflex_boosted_dt

WORLD_DT_STEPS = (apply_reflex_boosted_dt,)
PLAYER_DEATH_HOOKS = (apply_final_revenge_on_player_death,)




from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import pyray as rl

from grim.color import RGBA
from grim.geom import Vec2
from grim.math import clamp

from ..effects_atlas import EFFECT_ID_ATLAS_TABLE_BY_ID, EffectId, SIZE_CODE_GRID

if TYPE_CHECKING:
    from ..projectiles import SecondaryProjectile
    from .world_renderer import WorldRenderer


@dataclass(frozen=True, slots=True)
class SecondaryProjectileDrawCtx:
    renderer: WorldRenderer
    proj: SecondaryProjectile
    proj_type: int
    screen_pos: Vec2
    angle: float
    scale: float
    alpha: float


def _draw_secondary_rocket_like(ctx: SecondaryProjectileDrawCtx) -> bool:
    renderer = ctx.renderer
    proj_type = int(ctx.proj_type)
    if proj_type not in (1, 2, 4):
        return False

    texture = renderer.projs_texture
    if texture is None:
        return False

    cell_w = texture.width / 4.0
    if cell_w <= 1e-6:
        return True

    alpha = ctx.alpha
    base_alpha = clamp(alpha * 0.9, 0.0, 1.0)
    base_tint = RGBA(0.8, 0.8, 0.8, base_alpha).to_rl()
    base_size = 14.0
    if proj_type == 2:
        base_size = 10.0
    elif proj_type == 4:
        base_size = 8.0
    sprite_scale = (base_size * ctx.scale) / cell_w

    fx_detail_1 = bool(renderer.config.data.get("fx_detail_1", 0)) if renderer.config is not None else True
    if fx_detail_1 and renderer.particles_texture is not None:
        particles_texture = renderer.particles_texture
        atlas = EFFECT_ID_ATLAS_TABLE_BY_ID.get(int(EffectId.GLOW))
        if atlas is not None:
            grid = SIZE_CODE_GRID.get(int(atlas.size_code))
            if grid:
                particle_cell_w = particles_texture.width / grid
                particle_cell_h = particles_texture.height / grid
                frame = int(atlas.frame)
                col = frame % grid
                row = frame // grid
                src = rl.Rectangle(
                    particle_cell_w * col,
                    particle_cell_h * row,
                    max(0.0, particle_cell_w - 2.0),
                    max(0.0, particle_cell_h - 2.0),
                )

                direction = Vec2.from_heading(ctx.angle)

                scale = ctx.scale

                def _draw_rocket_fx(
                    *,
                    size: float,
                    offset: float,
                    rgba: RGBA,
                ) -> None:
                    fx_alpha = rgba.a
                    if fx_alpha <= 1e-3:
                        return
                    tint = rgba.to_rl()
                    fx_pos = ctx.screen_pos - direction * (offset * scale)
                    dst_size = size * scale
                    dst = rl.Rectangle(fx_pos.x, fx_pos.y, dst_size, dst_size)
                    origin = rl.Vector2(dst_size * 0.5, dst_size * 0.5)
                    rl.draw_texture_pro(particles_texture, src, dst, origin, 0.0, tint)

                rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)
                # Large bloom around the rocket (effect_id=0x0D).
                _draw_rocket_fx(size=140.0, offset=5.0, rgba=RGBA(1.0, 1.0, 1.0, alpha * 0.48))

                if proj_type == 4:
                    _draw_rocket_fx(size=30.0, offset=9.0, rgba=RGBA(0.7, 0.7, 1.0, alpha * 0.158))
                elif proj_type == 2:
                    _draw_rocket_fx(size=40.0, offset=9.0, rgba=RGBA(1.0, 1.0, 1.0, alpha * 0.58))
                else:
                    _draw_rocket_fx(size=60.0, offset=9.0, rgba=RGBA(1.0, 1.0, 1.0, alpha * 0.68))

                rl.end_blend_mode()

    renderer._draw_atlas_sprite(
        texture,
        grid=4,
        frame=3,
        pos=ctx.screen_pos,
        scale=sprite_scale,
        rotation_rad=ctx.angle,
        tint=base_tint,
    )
    return True


def _draw_secondary_type4_fallback(ctx: SecondaryProjectileDrawCtx) -> bool:
    if int(ctx.proj_type) != 4:
        return False
    rl.draw_circle(
        int(ctx.screen_pos.x),
        int(ctx.screen_pos.y),
        max(1.0, 12.0 * ctx.scale),
        rl.Color(200, 120, 255, int(255 * ctx.alpha + 0.5)),
    )
    return True


def _draw_secondary_detonation(ctx: SecondaryProjectileDrawCtx) -> bool:
    renderer = ctx.renderer
    if int(ctx.proj_type) != 3:
        return False

    # Secondary projectile detonation visuals (secondary_projectile_update + render).
    t = clamp(float(getattr(ctx.proj, "detonation_t", 0.0)), 0.0, 1.0)
    det_scale = float(getattr(ctx.proj, "detonation_scale", 1.0))
    fade = (1.0 - t) * ctx.alpha
    if fade <= 1e-3 or det_scale <= 1e-6:
        return True

    scale = ctx.scale
    particles_texture = renderer.particles_texture
    if particles_texture is None:
        radius = det_scale * t * 80.0
        alpha_byte = int(clamp((1.0 - t) * 180.0 * ctx.alpha, 0.0, 255.0) + 0.5)
        color = rl.Color(255, 180, 100, alpha_byte)
        rl.draw_circle_lines(int(ctx.screen_pos.x), int(ctx.screen_pos.y), max(1.0, radius * scale), color)
        return True

    atlas = EFFECT_ID_ATLAS_TABLE_BY_ID.get(int(EffectId.GLOW))
    if atlas is None:
        return True
    grid = SIZE_CODE_GRID.get(int(atlas.size_code))
    if not grid:
        return True
    frame = int(atlas.frame)
    col = frame % grid
    row = frame // grid
    cell_w = particles_texture.width / grid
    cell_h = particles_texture.height / grid
    src = rl.Rectangle(
        cell_w * col,
        cell_h * row,
        max(0.0, cell_w - 2.0),
        max(0.0, cell_h - 2.0),
    )

    def _draw_detonation_quad(*, size: float, alpha_mul: float) -> None:
        a = fade * alpha_mul
        if a <= 1e-3:
            return
        dst_size = size * scale
        if dst_size <= 1e-3:
            return
        tint = RGBA(1.0, 0.6, 0.1, a).to_rl()
        dst = rl.Rectangle(ctx.screen_pos.x, ctx.screen_pos.y, dst_size, dst_size)
        origin = rl.Vector2(dst_size * 0.5, dst_size * 0.5)
        rl.draw_texture_pro(particles_texture, src, dst, origin, 0.0, tint)

    rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)
    _draw_detonation_quad(size=det_scale * t * 64.0, alpha_mul=1.0)
    _draw_detonation_quad(size=det_scale * t * 200.0, alpha_mul=0.3)
    rl.end_blend_mode()
    return True


SECONDARY_PROJECTILE_DRAW_HANDLERS = (
    _draw_secondary_rocket_like,
    _draw_secondary_type4_fallback,
    _draw_secondary_detonation,
)


def draw_secondary_projectile_from_registry(ctx: SecondaryProjectileDrawCtx) -> bool:
    for handler in SECONDARY_PROJECTILE_DRAW_HANDLERS:
        if handler(ctx):
            return True
    return False



from __future__ import annotations

"""Creature AI helpers.

Ported from `creature_update_all` (`FUN_00426220`).
"""

from dataclasses import dataclass
import math
from typing import Callable, Protocol, Sequence

from grim.geom import Vec2

from ..math_parity import NATIVE_PI, f32, f32_vec2, heading_from_delta_f32
from .spawn import CreatureFlags

__all__ = [
    "CreatureAIUpdate",
    "creature_ai7_tick_link_timer",
    "creature_ai_update_target",
]


class PositionLike(Protocol):
    pos: Vec2


class CreatureLinkLike(PositionLike, Protocol):
    hp: float


class CreatureAIStateLike(CreatureLinkLike, Protocol):
    flags: CreatureFlags
    ai_mode: int
    link_index: int
    target_offset: Vec2 | None
    phase_seed: float
    orbit_angle: float
    orbit_radius: float
    heading: float

    target: Vec2
    target_heading: float
    force_target: int


@dataclass(frozen=True, slots=True)
class CreatureAIUpdate:
    move_scale: float
    self_damage: float | None = None


def creature_ai7_tick_link_timer(creature: CreatureAIStateLike, *, dt_ms: int, rand: Callable[[], int]) -> None:
    """Update AI7's link-index timer behavior (flag 0x80).

    In the original, this runs regardless of the current ai_mode; when the timer
    flips from negative to non-negative, ai_mode is forced to 7 for a short hold.
    """

    if not (creature.flags & CreatureFlags.AI7_LINK_TIMER):
        return

    if creature.link_index < 0:
        creature.link_index += dt_ms
        if creature.link_index >= 0:
            creature.ai_mode = 7
            creature.link_index = (rand() & 0x1FF) + 500
        return

    creature.link_index -= dt_ms
    if creature.link_index < 1:
        creature.link_index = -700 - (rand() & 0x3FF)


def resolve_live_link(creatures: Sequence[CreatureLinkLike], link_index: int) -> CreatureLinkLike | None:
    if 0 <= link_index < len(creatures) and creatures[link_index].hp > 0.0:
        return creatures[link_index]
    return None


def _distance_f32(a: Vec2, b: Vec2) -> float:
    dx = f32(float(b.x) - float(a.x))
    dy = f32(float(b.y) - float(a.y))
    dist_sq = f32(f32(float(dx) * float(dx)) + f32(float(dy) * float(dy)))
    return f32(math.sqrt(float(dist_sq)))


def _orbit_target_f32(*, player_pos: Vec2, orbit_phase: float, dist: float, scale: float) -> Vec2:
    orbit_dist = f32(f32(float(dist)) * f32(float(scale)))
    phase = f32(float(orbit_phase))
    px = f32(float(player_pos.x))
    py = f32(float(player_pos.y))
    orbit_x = f32(math.cos(float(phase)))
    orbit_y = f32(math.sin(float(phase)))
    return Vec2(
        f32(f32(float(orbit_x) * float(orbit_dist)) + px),
        f32(f32(float(orbit_y) * float(orbit_dist)) + py),
    )


def _link_target_f32(*, link_pos: Vec2, offset: Vec2) -> Vec2:
    return Vec2(
        f32(float(link_pos.x) + float(offset.x)),
        f32(float(link_pos.y) + float(offset.y)),
    )


def creature_ai_update_target(
    creature: CreatureAIStateLike,
    *,
    player_pos: Vec2,
    creatures: Sequence[CreatureLinkLike],
    dt: float,
) -> CreatureAIUpdate:
    """Compute the target position + heading for one creature.

    Updates:
    - `target`
    - `target_heading`
    - `force_target`
    - `ai_mode` (may reset to 0 in some modes)
    - `orbit_radius` (AI7 non-link timer uses it as a countdown)
    """

    dist_to_player = _distance_f32(creature.pos, player_pos)
    orbit_phase = f32(f32(float(int(creature.phase_seed)) * f32(3.7)) * NATIVE_PI)
    move_scale = 1.0
    self_damage: float | None = None

    creature.force_target = 0

    ai_mode = creature.ai_mode
    if ai_mode == 0:
        if dist_to_player > 800.0:
            creature.target = f32_vec2(player_pos)
        else:
            creature.target = _orbit_target_f32(
                player_pos=player_pos,
                orbit_phase=orbit_phase,
                dist=dist_to_player,
                scale=0.85,
            )
    elif ai_mode == 8:
        creature.target = _orbit_target_f32(
            player_pos=player_pos,
            orbit_phase=orbit_phase,
            dist=dist_to_player,
            scale=0.9,
        )
    elif ai_mode == 1:
        if dist_to_player > 800.0:
            creature.target = f32_vec2(player_pos)
        else:
            creature.target = _orbit_target_f32(
                player_pos=player_pos,
                orbit_phase=orbit_phase,
                dist=dist_to_player,
                scale=0.55,
            )
    elif ai_mode == 3:
        link = resolve_live_link(creatures, creature.link_index)
        if link is not None:
            creature.target = _link_target_f32(link_pos=link.pos, offset=(creature.target_offset or Vec2()))
        else:
            creature.ai_mode = 0
    elif ai_mode == 5:
        link = resolve_live_link(creatures, creature.link_index)
        if link is not None:
            creature.target = _link_target_f32(link_pos=link.pos, offset=(creature.target_offset or Vec2()))
            dist_to_target = _distance_f32(creature.pos, creature.target)
            if dist_to_target <= 64.0:
                move_scale = f32(dist_to_target * 0.015625)
        else:
            creature.ai_mode = 0
            self_damage = 1000.0

    ai_mode = creature.ai_mode
    if ai_mode == 4:
        link = resolve_live_link(creatures, creature.link_index)
        if link is None:
            creature.ai_mode = 0
            self_damage = 1000.0
        elif dist_to_player > 800.0:
            creature.target = f32_vec2(player_pos)
        else:
            creature.target = _orbit_target_f32(
                player_pos=player_pos,
                orbit_phase=orbit_phase,
                dist=dist_to_player,
                scale=0.85,
            )
    elif ai_mode == 7:
        if (creature.flags & CreatureFlags.AI7_LINK_TIMER) and creature.link_index > 0:
            creature.target = f32_vec2(creature.pos)
        elif not (creature.flags & CreatureFlags.AI7_LINK_TIMER) and creature.orbit_radius > 0.0:
            creature.target = f32_vec2(creature.pos)
            creature.orbit_radius = f32(float(creature.orbit_radius) - float(dt))
        else:
            creature.ai_mode = 0
    elif ai_mode == 6:
        link = resolve_live_link(creatures, creature.link_index)
        if link is None:
            creature.ai_mode = 0
        else:
            angle = float(creature.orbit_angle) + float(creature.heading)
            orbit_radius = float(creature.orbit_radius)
            creature.target = Vec2(
                f32(math.cos(angle) * orbit_radius + float(link.pos.x)),
                f32(math.sin(angle) * orbit_radius + float(link.pos.y)),
            )

    dist_to_target = _distance_f32(creature.pos, creature.target)
    if dist_to_target < 40.0 or dist_to_target > 400.0:
        creature.force_target = 1

    if creature.force_target or creature.ai_mode == 2:
        creature.target = f32_vec2(player_pos)

    creature.target_heading = heading_from_delta_f32(
        dx=float(creature.target.x) - float(creature.pos.x),
        dy=float(creature.target.y) - float(creature.pos.y),
    )
    return CreatureAIUpdate(move_scale=f32(move_scale), self_damage=self_damage)
